import json
import logging
import math
import os
import re
import urllib.request
from collections.abc import Mapping
from pathlib import Path
import hashlib

import requests

from ibllib.misc import pprint, print_progress

_logger = logging.getLogger('ibllib')


class _PaginatedResponse(Mapping):
    """
    This class allows to emulate a list from a paginated response.
    Provides cache functionality
    PaginatedResponse(alyx, response)
    """
    def __init__(self, alyx, rep):
        self.alyx = alyx
        self.count = rep['count']
        self.limit = len(rep['results'])
        # warning: the offset and limit filters are not necessarily the last ones
        lquery = [q for q in rep['next'].split('&')
                  if not (q.startswith('offset=') or q.startswith('limit='))]
        self.query = '&'.join(lquery)
        # init the cache, list with None with count size
        self._cache = [None for _ in range(self.count)]
        # fill the cache with results of the query
        for i in range(self.limit):
            self._cache[i] = rep['results'][i]

    def __len__(self):
        return self.count

    def __getitem__(self, item):
        if self._cache[item] is None:
            offset = self.limit * math.floor(item / self.limit)
            query = f'{self.query}&limit={self.limit}&offset={offset}'
            res = self.alyx._generic_request(requests.get, query)
            for i, r in enumerate(res['results']):
                self._cache[i + offset] = res['results'][i]
        return self._cache[item]

    def __iter__(self):
        for i in range(self.count):
            yield self.__getitem__(i)


def http_download_file_list(links_to_file_list, **kwargs):
    """
    Downloads a list of files from the flat Iron from a list of links.
    Same options behaviour as http_download_file

    :param links_to_file_list: list of http links to files.
    :type links_to_file_list: list

    :return: (list) a list of the local full path of the downloaded files.
    """
    file_names_list = []
    for link_str in links_to_file_list:
        file_names_list.append(http_download_file(link_str, **kwargs))
    return file_names_list


def http_download_file(full_link_to_file, *, clobber=False, offline=False,
                       username='', password='', cache_dir='', return_md5=False):
    """
    :param full_link_to_file: http link to the file.
    :type full_link_to_file: str
    :param clobber: [False] If True, force overwrite the existing file.
    :type clobber: bool
    :param username: [''] authentication for password protected file server.
    :type username: str
    :param password: [''] authentication for password protected file server.
    :type password: str
    :param cache_dir: [''] directory in which files are cached; defaults to user's
     Download directory.
    :type cache_dir: str

    :return: (str) a list of the local full path of the downloaded files.
    """
    from ibllib.io import hashfile
    if not full_link_to_file:
        return ''

    # default cache directory is the home dir
    if not cache_dir:
        cache_dir = str(Path.home().joinpath("Downloads"))

    # This is the local file name
    file_name = str(cache_dir) + os.sep + os.path.basename(full_link_to_file)

    # do not overwrite an existing file unless specified
    if not clobber and os.path.exists(file_name):
        return (file_name, hashfile.md5(file_name)) if return_md5 else file_name
    elif offline:
        return (file_name, hashfile.md5(file_name)) if return_md5 else file_name

    # This should be the base url you wanted to access.
    baseurl = os.path.split(str(full_link_to_file))[0]

    # Create a password manager
    manager = urllib.request.HTTPPasswordMgrWithDefaultRealm()
    if (len(password) != 0) & (len(username) != 0):
        manager.add_password(None, baseurl, username, password)

    # Create an authentication handler using the password manager
    auth = urllib.request.HTTPBasicAuthHandler(manager)

    # Create an opener that will replace the default urlopen method on further calls
    opener = urllib.request.build_opener(auth)
    urllib.request.install_opener(opener)

    # Open the url and get the length
    try:
        u = urllib.request.urlopen(full_link_to_file)
    except urllib.error.HTTPError as e:
        _logger.error(f"{str(e)} {full_link_to_file}")
        raise e

    file_size = int(u.getheader('Content-length'))

    print(f"Downloading: {file_name} Bytes: {file_size}")
    file_size_dl = 0
    block_sz = 8192 * 64 * 8

    md5 = hashlib.md5()
    f = open(file_name, 'wb')
    while True:
        buffer = u.read(block_sz)
        if not buffer:
            break
        file_size_dl += len(buffer)
        f.write(buffer)
        if return_md5:
            md5.update(buffer)
        print_progress(file_size_dl, file_size, prefix='', suffix='')
    f.close()

    return (file_name, md5.hexdigest()) if return_md5 else file_name


def file_record_to_url(file_records, urls=[]):
    """
    Translate a Json dictionary to an usable http url for downlading files.

    :param file_records: json containing a 'data_url' field
    :type file_records: dict
    :param urls: a list of strings containing previous data_urls on which new urls
     will be appended
    :type urls: list

    :return: urls: (list) a list of strings representing full data urls
    """
    for fr in file_records:
        if fr['data_url'] is not None:
            urls.append(fr['data_url'])
    return urls


def dataset_record_to_url(dataset_record):
    """
    Extracts a list of files urls from a list of dataset queries.

    :param dataset_record: dataset Json from a rest request.
    :type dataset_record: list

    :return: (list) a list of strings representing files urls corresponding to the datasets records
    """
    urls = []
    if type(dataset_record) is dict:
        dataset_record = [dataset_record]
    for ds in dataset_record:
        urls = file_record_to_url(ds['file_records'], urls)
    return urls


class UniqueSingletons(type):
    _instances: list = []

    def __call__(cls, *args, **kwargs):
        # print('args', args, '\nkwargs', kwargs)
        for inst in UniqueSingletons._instances:
            if cls in inst and inst.get(cls, None).get('args') == (args, kwargs):
                return inst[cls].get('instance')

        new_instance = super(UniqueSingletons, cls).__call__(*args, **kwargs)
        # Optional rerun of constructor
        # new_instance.__init__(*args, **kwargs)
        new_instance_record = {
            cls: {'args': (args, kwargs), 'instance': new_instance}
        }
        UniqueSingletons._instances.append(new_instance_record)

        return new_instance


class AlyxClient(metaclass=UniqueSingletons):
    """
    Class that implements simple GET/POST wrappers for the Alyx REST API
    http://alyx.readthedocs.io/en/latest/api.html
    """
    _token = ''  # These class params are not being used!!
    _headers = ''
    _rest_schemes = ''

    def __init__(self, **kwargs):
        """
        Create a client instance that allows to GET and POST to the Alyx server
        For oneibl, constructor attempts to authenticate with credentials in params.py
        For standalone cases, AlyxClient(username='', password='', base_url='')

        :param username: Alyx database user
        :type username: str
        :param password: Alyx database password
        :type password: str
        :param base_url: Alyx server address, including port and protocol
        :type base_url: str
        """
        self.authenticate(**kwargs)
        self._headers['Accept'] = 'application/coreapi+json'
        self._rest_schemes = self.get('/docs')
        # the mixed accept application may cause errors sometimes, only necessary for the docs
        self._headers['Accept'] = 'application/json'
        self._obj_id = id(self)

    def _generic_request(self, reqfunction, rest_query, data=None):
        # if the data is a dictionary, it has to be converted to json text
        if isinstance(data, dict) or isinstance(data, list):
            data = json.dumps(data)
        # makes sure the base url is the one from the instance
        rest_query = rest_query.replace(self._base_url, '')
        if not rest_query.startswith('/'):
            rest_query = '/' + rest_query
        _logger.debug(self._base_url + rest_query)
        r = reqfunction(self._base_url + rest_query, stream=True, headers=self._headers, data=data)
        if r and r.status_code in (200, 201):
            return json.loads(r.text)
        elif r and r.status_code == 204:
            return
        else:
            _logger.error(self._base_url + rest_query)
            _logger.error(r.text)
            raise(requests.HTTPError(r))

    def authenticate(self, username='', password='', base_url=''):
        """
        Gets a security token from the Alyx REST API to create requests headers.
        Credentials are in the params_secret_template.py file

        :param username: Alyx database user
        :type username: str
        :param password: Alyx database password
        :type password: str
        :param base_url: Alyx server address, including port and protocol
        :type base_url: str
        """
        self._base_url = base_url
        rep = requests.post(base_url + '/auth-token',
                            data=dict(username=username, password=password))
        # Assign token or raise exception on internal server error
        self._token = rep.json() if rep.ok else rep.raise_for_status()
        if not (list(self._token.keys()) == ['token']):
            _logger.error(rep)
            raise Exception('Alyx authentication error. Check your credentials')
        self._headers = {
            'Authorization': 'Token {}'.format(list(self._token.values())[0]),
            'Accept': 'application/json',
            'Content-Type': 'application/json'}

    def delete(self, rest_query):
        """
        Sends a DELETE request to the Alyx server. Will raise an exception on any status_code
        other than 200, 201.

        :param rest_query: examples:
         '/weighings/c617562d-c107-432e-a8ee-682c17f9e698'
         'https://test.alyx.internationalbrainlab.org/weighings/c617562d-c107-432e-a8ee-682c17f9e698'.
        :type rest_query: str

        :return: (dict/list) json interpreted dictionary from response
        """
        return self._generic_request(requests.delete, rest_query)

    def get(self, rest_query):
        """
        Sends a GET request to the Alyx server. Will raise an exception on any status_code
        other than 200, 201.
        For the dictionary contents and list of endpoints, refer to:
        https://alyx.internationalbrainlab.org/docs

        :param rest_query: example: '/sessions?user=Hamish'.
        :type rest_query: str

        :return: (dict/list) json interpreted dictionary from response
        """
        rep = self._generic_request(requests.get, rest_query)
        _logger.debug(rest_query)
        if isinstance(rep, dict) and list(rep.keys()) == ['count', 'next', 'previous', 'results']:
            if len(rep['results']) < rep['count']:
                rep = _PaginatedResponse(self, rep)
            else:
                rep = rep['results']
        return rep

    def patch(self, rest_query, data=None):
        """
        Sends a PATCH request to the Alyx server.
        For the dictionary contents, refer to:
        https://alyx.internationalbrainlab.org/docs

        :param rest_query: (required)the endpoint as full or relative URL
        :type rest_query: str
        :param data: json encoded string or dictionary
        :type data: None, dict or str

        :return: response object
        """
        if isinstance(data, dict):
            data = json.dumps(data)
        return self._generic_request(requests.patch, rest_query, data=data)

    def post(self, rest_query, data=None):
        """
        Sends a POST request to the Alyx server.
        For the dictionary contents, refer to:
        https://alyx.internationalbrainlab.org/docs

        :param rest_query: (required)the endpoint as full or relative URL
        :type rest_query: str
        :param data: dictionary or json encoded string
        :type data: None, dict or str

        :return: response object
        """
        return self._generic_request(requests.post, rest_query, data=data)

    def put(self, rest_query, data=None):
        """
        Sends a PUT request to the Alyx server.
        For the dictionary contents, refer to:
        https://alyx.internationalbrainlab.org/docs

        :param rest_query: (required)the endpoint as full or relative URL
        :type rest_query: str
        :param data: dictionary or json encoded string
        :type data: None, dict or str

        :return: response object
        """
        return self._generic_request(requests.put, rest_query, data=data)

    def rest(self, url=None, action=None, id=None, data=None, **kwargs):
        """
        alyx_client.rest(): lists endpoints
        alyx_client.rest(endpoint): lists actions for endpoint
        alyx_client.rest(endpoint, action): lists fields and URL

        Example with a rest endpoint with all actions

        >>> alyx.client.rest('subjects', 'list')
            alyx.client.rest('subjects', 'list', field_filter1='filterval')
            alyx.client.rest('subjects', 'create', data=sub_dict)
            alyx.client.rest('subjects', 'read', id='nickname')
            alyx.client.rest('subjects', 'update', id='nickname', data=sub_dict)
            alyx.client.rest('subjects', 'partial_update', id='nickname', data=sub_dict)
            alyx.client.rest('subjects', 'delete', id='nickname')

        :param url: endpoint name
        :param action: 'list', 'create', 'read', 'update', 'partial_update', 'delete'
        :param id: lookup string for actions 'read', 'update', 'partial_update', and 'delete'
        :param data: data dictionary for actions 'update', 'partial_update' and 'create'
        :param ``**kwargs``: filter as per the REST documentation
        :return: list of queried dicts ('list') or dict (other actions)
        """
        # if endpoint is None, list available endpoints
        if not url:
            pprint([k for k in self._rest_schemes.keys() if not k.startswith('_') and k])
            return
        # remove beginning slash if any
        if url.startswith('/'):
            url = url[1:]
        # and split to the next slash or question mark
        endpoint = re.findall("^/*[^?/]*", url)[0].replace('/', '')
        # make sure the queryied endpoint exists, if not throw an informative error
        if endpoint not in self._rest_schemes.keys():
            av = [k for k in self._rest_schemes.keys() if not k.startswith('_') and k]
            raise ValueError('REST endpoint "' + endpoint + '" does not exist. Available ' +
                             'endpoints are \n       ' + '\n       '.join(av))
        endpoint_scheme = self._rest_schemes[endpoint]
        # on a filter request, override the default action parameter
        if '?' in url:
            action = 'list'
        # if action is None, list available actions for the required endpoint
        if not action:
            pprint(list(endpoint_scheme.keys()))
            return
        # make sure the the desired action exists, if not throw an informative error
        if action not in endpoint_scheme.keys():
            raise ValueError('Action "' + action + '" for REST endpoint "' + endpoint + '" does ' +
                             'not exist. Available actions are: ' +
                             '\n       ' + '\n       '.join(endpoint_scheme.keys()))
        # the actions below require an id in the URL, warn and help the user
        if action in ['read', 'update', 'partial_update', 'delete'] and not id:
            _logger.warning('REST action "' + action + '" requires an ID in the URL: ' +
                            endpoint_scheme[action]['url'])
            return
        # the actions below require a data dictionary, warn and help the user with fields list
        if action in ['create', 'update', 'partial_update'] and not data:
            pprint(endpoint_scheme[action]['fields'])
            for act in endpoint_scheme[action]['fields']:
                print("'" + act['name'] + "': ...,")
            _logger.warning('REST action "' + action + '" requires a data dict with above keys')
            return

        if action == 'list':
            # list doesn't require id nor
            assert(endpoint_scheme[action]['action'] == 'get')
            # add to url data if it is a string
            if id:
                url = url + id
            # otherwise, look for a dictionary of filter terms
            elif kwargs:
                url += '?'
                for k in kwargs.keys():
                    if isinstance(kwargs[k], str):
                        query = kwargs[k]
                    elif isinstance(kwargs[k], list):
                        query = ','.join(kwargs[k])
                    else:
                        query = str(kwargs[k])
                    url = url + f"&{k}=" + query
            return self.get('/' + url)
        if action == 'read':
            assert(endpoint_scheme[action]['action'] == 'get')
            return self.get('/' + endpoint + '/' + id.split('/')[-1])
        elif action == 'create':
            assert(endpoint_scheme[action]['action'] == 'post')
            return self.post('/' + endpoint, data=data)
        elif action == 'delete':
            assert(endpoint_scheme[action]['action'] == 'delete')
            return self.delete('/' + endpoint + '/' + id.split('/')[-1])
        elif action == 'partial_update':
            assert(endpoint_scheme[action]['action'] == 'patch')
            return self.patch('/' + endpoint + '/' + id.split('/')[-1], data=data)
        elif action == 'update':
            assert(endpoint_scheme[action]['action'] == 'put')
            return self.put('/' + endpoint + '/' + id.split('/')[-1], data=data)

    # JSON field interface convenience methods
    def _check_inputs(self, endpoint: str) -> None:
        # make sure the queryied endpoint exists, if not throw an informative error
        if endpoint not in self._rest_schemes.keys():
            av = [k for k in self._rest_schemes.keys() if not k.startswith('_') and k]
            raise ValueError('REST endpoint "' + endpoint + '" does not exist. Available ' +
                             'endpoints are \n       ' + '\n       '.join(av))
        return

    def json_field_write(
        self,
        endpoint: str = None,
        uuid: str = None,
        field_name: str = None,
        data: dict = None
    ) -> dict:
        """json_field_write [summary]
        Write data to WILL NOT CHECK IF DATA EXISTS
        NOTE: Destructive write!

        :param endpoint: Valid alyx endpoint, defaults to None
        :type endpoint: str, optional
        :param uuid: uuid or lookup name for endpoint
        :type uuid: str, optional
        :param field_name: Valid json field name, defaults to None
        :type field_name: str, optional
        :param data: data to write to json field, defaults to None
        :type data: dict, optional
        :return: Written data dict
        :rtype: dict
        """
        self._check_inputs(endpoint)
        # Prepare data to patch
        patch_dict = {field_name: data}
        # Upload new extended_qc to session
        ret = self.rest(endpoint, "partial_update", id=uuid, data=patch_dict)
        return ret[field_name]

    def json_field_update(
        self,
        endpoint: str = None,
        uuid: str = None,
        field_name: str = 'json',
        data: dict = None
    ) -> dict:
        """json_field_update
        Non destructive update of json field of endpoint for object
        Will update the field_name of the object with pk = uuid of given endpoint
        If data has keys with the same name of existing keys it will squash the old
        values (uses the dict.update() method)

        Example:
        one.alyx.json_field_update("sessions", "eid_str", "extended_qc" {"key": value})

        :param endpoint: endpoint to hit
        :type endpoint: str
        :param uuid: uuid or lookup name of object
        :type uuid: str
        :param field_name: name of the json field
        :type field_name: str
        :param data: dictionary with fields to be updated
        :type data: dict
        :return: new patched json field contents
        :rtype: dict
        """
        self._check_inputs(endpoint)
        # Load current json field contents
        current = self.rest(endpoint, "read", id=uuid)[field_name]
        if current is None:
            current = {}

        if not isinstance(current, dict):
            _logger.warning(
                f"Current json field {field_name} does not contains a dict, aborting update"
            )
            return current

        # Patch current dict with new data
        current.update(data)
        # Prepare data to patch
        patch_dict = {field_name: current}
        # Upload new extended_qc to session
        ret = self.rest(endpoint, "partial_update", id=uuid, data=patch_dict)
        return ret[field_name]

    def json_field_remove_key(
        self,
        endpoint: str = None,
        uuid: str = None,
        field_name: str = 'json',
        key: str = None
    ) -> dict:
        """json_field_remove_key
        Will remove inputted key from json field dict and reupload it to Alyx.
        Needs endpoint, uuid and json field name

        :param endpoint: endpoint to hit, defaults to None
        :type endpoint: str, optional
        :param uuid: uuid or lookup name for endpoint
        :type uuid: str, optional
        :param field_name: json field name of object, defaults to None
        :type field_name: str, optional
        :param key: key name of dictionary inside object, defaults to None
        :type key: str, optional
        :return: returns new content of json field
        :rtype: dict
        """
        self._check_inputs(endpoint)
        current = self.rest(endpoint, "read", id=uuid)[field_name]
        # If no contents, cannot remove key, return
        if current is None:
            return current
        # if contents are not dict, cannot remove key, return contents
        if isinstance(current, str):
            _logger.warning(f"Cannot remove key {key} content of json field is of type str")
            return current
        # If key not present in contents of json field cannot remove key, return contents
        if current.get(key, None) is None:
            _logger.warning(
                f"{key}: Key not found in endpoint {endpoint} field {field_name}"
            )
            return current
        _logger.info(f"Removing key from dict: '{key}'")
        current.pop(key)
        # Re-write contents without removed key
        written = self.json_field_write(
            endpoint=endpoint, uuid=uuid, field_name=field_name, data=current
        )
        return written

    def json_field_delete(
        self, endpoint: str = None, uuid: str = None, field_name: str = None
    ) -> None:
        self._check_inputs(endpoint)
        _ = self.rest(endpoint, "partial_update", id=uuid, data={field_name: None})
        return _[field_name]



import json
import logging
from pathlib import Path
import re

import numpy as np

import mtscomp
from brainbox.core import Bunch
from ibllib.ephys import neuropixel as neuropixel
from ibllib.io import hashfile

SAMPLE_SIZE = 2  # int16
DEFAULT_BATCH_SIZE = 1e6
_logger = logging.getLogger('ibllib')


class Reader:
    """
    Class for SpikeGLX reading purposes
    Some format description was found looking at the Matlab SDK here
    https://github.com/billkarsh/SpikeGLX/blob/master/MATLAB-SDK/DemoReadSGLXData.m
    """
    def __init__(self, sglx_file):
        self.file_bin = Path(sglx_file)
        self.nbytes = self.file_bin.stat().st_size
        file_meta_data = Path(sglx_file).with_suffix('.meta')
        if not file_meta_data.exists():
            self.file_meta_data = None
            self.meta = None
            self.channel_conversion_sample2v = 1
            _logger.warning(str(sglx_file) + " : no metadata file found. Very limited support")
            return
        # normal case we continue reading and interpreting the metadata file
        self.file_meta_data = file_meta_data
        self.meta = read_meta_data(file_meta_data)
        self.channel_conversion_sample2v = _conversion_sample2v_from_meta(self.meta)
        # if we are not looking at a compressed file, use a memmap, otherwise instantiate mtscomp
        if self.is_mtscomp:
            self._raw = mtscomp.Reader()
            self._raw.open(self.file_bin, self.file_bin.with_suffix('.ch'))
        else:
            if self.nc * self.ns * 2 != self.nbytes:
                ftsec = self.file_bin.stat().st_size / 2 / self.nc / self.fs
                _logger.warning(f"{sglx_file} : meta data and filesize do not checkout\n"
                                f"File size: expected {self.meta['fileSizeBytes']},"
                                f" actual {self.file_bin.stat().st_size}\n"
                                f"File duration: expected {self.meta['fileTimeSecs']},"
                                f" actual {ftsec}\n"
                                f"Will attempt to fudge the meta-data information.")
                self.meta['fileTimeSecs'] = ftsec
            self._raw = np.memmap(sglx_file, dtype='int16', mode='r', shape=(self.ns, self.nc))

    def __getitem__(self, item):
        if isinstance(item, int) or isinstance(item, slice):
            return self.read(nsel=item, sync=False)
        elif len(item) == 2:
            return self.read(nsel=item[0], csel=item[1], sync=False)

    @property
    def shape(self):
        return self.ns, self.nc

    @property
    def is_mtscomp(self):
        return 'cbin' in self.file_bin.suffix

    @property
    def version(self):
        """:return: """
        if not self.meta:
            return None
        return _get_neuropixel_version_from_meta(self.meta)

    @property
    def type(self):
        """:return: ap, lf or nidq. Useful to index dictionaries """
        if not self.meta:
            return 0
        return _get_type_from_meta(self.meta)

    @property
    def fs(self):
        """ :return: sampling frequency (Hz) """
        if not self.meta:
            return 1
        return _get_fs_from_meta(self.meta)

    @property
    def nc(self):
        """ :return: number of channels """
        if not self.meta:
            return
        return _get_nchannels_from_meta(self.meta)

    @property
    def ns(self):
        """ :return: number of samples """
        if not self.meta:
            return
        return int(np.round(self.meta.get('fileTimeSecs') * self.fs))

    def read(self, nsel=slice(0, 10000), csel=slice(None), sync=True):
        """
        Read from slices or indexes
        :param slice_n: slice or sample indices
        :param slice_c: slice or channel indices
        :return: float32 array
        """
        darray = np.float32(self._raw[nsel, csel])
        darray *= self.channel_conversion_sample2v[self.type][csel]
        if sync:
            return darray, self.read_sync(nsel)
        else:
            return darray

    def read_samples(self, first_sample=0, last_sample=10000, channels=None):
        """
        reads all channels from first_sample to last_sample, following numpy slicing convention
        sglx.read_samples(first=0, last=100) would be equivalent to slicing the array D
        D[:,0:100] where the last axis represent time and the first channels.

         :param first_sample: first sample to be read, python slice-wise
         :param last_sample:  last sample to be read, python slice-wise
         :param channels: slice or numpy array of indices
         :return: numpy array of int16
        """
        if channels is None:
            channels = slice(None)
        return self.read(slice(first_sample, last_sample), channels)

    def read_sync_digital(self, _slice=slice(0, 10000)):
        """
        Reads only the digital sync trace at specified samples using slicing syntax
        >>> sync_samples = sr.read_sync_digital(slice(0,10000))
        """
        if not self.meta:
            _logger.warning('Sync trace not labeled in metadata. Assuming last trace')
        return split_sync(self._raw[_slice, _get_sync_trace_indices_from_meta(self.meta)])

    def read_sync_analog(self, _slice=slice(0, 10000)):
        """
        Reads only the analog sync traces at specified samples using slicing syntax
        >>> sync_samples = sr.read_sync_analog(slice(0,10000))
        """
        if not self.meta:
            return
        csel = _get_analog_sync_trace_indices_from_meta(self.meta)
        if not csel:
            return
        else:
            return self.read(nsel=_slice, csel=csel, sync=False)

    def read_sync(self, _slice=slice(0, 10000), threshold=1.2):
        """
        Reads all sync trace. Convert analog to digital with selected threshold and append to array
        :param _slice: samples slice
        :param threshold: (V) threshold for front detection, defaults to 1.2 V
        :return: int8 array
        """
        digital = self.read_sync_digital(_slice)
        analog = self.read_sync_analog(_slice)
        if analog is None:
            return digital
        analog[np.where(analog < threshold)] = 0
        analog[np.where(analog >= threshold)] = 1
        return np.concatenate((digital, np.int8(analog)), axis=1)

    def compress_file(self, keep_original=True, **kwargs):
        """
        Compresses
        :param keep_original: defaults True. If False, the original uncompressed file is deleted
         and the current spikeglx.Reader object is modified in place
        :param kwargs:
        :return: pathlib.Path of the compressed *.cbin file
        """
        file_tmp = self.file_bin.with_suffix('.cbin_tmp')
        assert not self.is_mtscomp
        mtscomp.compress(self.file_bin,
                         out=file_tmp,
                         outmeta=self.file_bin.with_suffix('.ch'),
                         sample_rate=self.fs,
                         n_channels=self.nc,
                         dtype=np.int16,
                         **kwargs)
        file_out = file_tmp.with_suffix('.cbin')
        file_tmp.rename(file_out)
        if not keep_original:
            self.file_bin.unlink()
            self.file_bin = file_out
        return file_out

    def decompress_file(self, keep_original=True, **kwargs):
        """
        Decompresses a mtscomp file
        :param keep_original: defaults True. If False, the original compressed file (input)
        is deleted and the current spikeglx.Reader object is modified in place
        NB: This is not equivalent to overwrite (which replaces the output file)
        :return: pathlib.Path of the decompressed *.bin file
        """
        if 'out' not in kwargs:
            kwargs['out'] = self.file_bin.with_suffix('.bin')
        assert self.is_mtscomp
        mtscomp.decompress(self.file_bin, self.file_bin.with_suffix('.ch'), **kwargs)
        if not keep_original:
            self.file_bin.unlink()
            self.file_bin.with_suffix('.ch').unlink()
            self.file_bin = kwargs['out']
        return kwargs['out']

    def verify_hash(self):
        """
        Computes SHA-1 hash and returns True if it matches metadata, False otherwise
        :return: boolean
        """
        if self.is_mtscomp:
            with open(self.file_bin.with_suffix('.ch')) as fid:
                mtscomp_params = json.load(fid)
            sm = mtscomp_params.get('sha1_compressed', None)
            if sm is None:
                _logger.warning("SHA1 hash is not implemented for compressed ephys. To check "
                                "the spikeglx acquisition hash, uncompress the file first !")
                return True
            sm = sm.upper()
        else:
            sm = self.meta.fileSHA1
        sc = hashfile.sha1(self.file_bin).upper()
        if sm == sc:
            log_func = _logger.info
        else:
            log_func = _logger.error
        log_func(f"SHA1 metadata: {sm}")
        log_func(f"SHA1 computed: {sc}")
        return sm == sc


def read(sglx_file, first_sample=0, last_sample=10000):
    """
    Function to read from a spikeglx binary file without instantiating the class.
    Gets the meta-data as well.

    >>> ibllib.io.spikeglx.read('/path/to/file.bin', first_sample=0, last_sample=1000)

    :param sglx_file: full path the the binary file to read
    :param first_sample: first sample to be read, python slice-wise
    :param last_sample: last sample to be read, python slice-wise
    :return: Data array, sync trace, meta-data
    """
    sglxr = Reader(sglx_file)
    D, sync = sglxr.read_samples(first_sample=first_sample, last_sample=last_sample)
    return D, sync, sglxr.meta


def read_meta_data(md_file):
    """
    Reads the spkike glx metadata file and parse in a dictionary
    Agnostic: does not make any assumption on the keys/content, it just parses key=values

    :param md_file: last sample to be read, python slice-wise
    :return: Data array, sync trace, meta-data
    """
    with open(md_file) as fid:
        md = fid.read()
    d = {}
    for a in md.splitlines():
        k, v = a.split('=')
        # if all numbers, try to interpret the string
        if v and re.fullmatch('[0-9,.]*', v) and v.count('.') < 2:
            v = [float(val) for val in v.split(',')]
            # scalars should not be nested
            if len(v) == 1:
                v = v[0]
        # tildes in keynames removed
        d[k.replace('~', '')] = v
    d['neuropixelVersion'] = _get_neuropixel_version_from_meta(d)
    d['serial'] = _get_serial_number_from_meta(d)
    return Bunch(d)


def _get_serial_number_from_meta(md):
    """
    Get neuropixel serial number from the metadata dictionary
    """
    # imProbeSN for 3A, imDatPrb_sn for 3B2, None for nidq 3B2
    serial = md.get('imProbeSN') or md.get('imDatPrb_sn')
    if serial:
        return int(serial)


def _get_neuropixel_version_from_meta(md):
    """
    Get neuropixel version tag (3A, 3B1, 3B2) from the metadata dictionary
    """
    if 'typeEnabled' in md.keys():
        return '3A'
    elif 'typeImEnabled' in md.keys() and 'typeNiEnabled' in md.keys():
        if 'imDatPrb_port' in md.keys() and 'imDatPrb_slot' in md.keys():
            return '3B2'
        else:
            return '3B1'


def _get_sync_trace_indices_from_meta(md):
    """
    Returns a list containing indices of the sync traces in the original array
    """
    typ = _get_type_from_meta(md)
    ntr = int(_get_nchannels_from_meta(md))
    if typ == 'nidq':
        nsync = int(md.get('snsMnMaXaDw')[-1])
    elif typ in ['lf', 'ap']:
        nsync = int(md.get('snsApLfSy')[2])
    return list(range(ntr - nsync, ntr))


def _get_analog_sync_trace_indices_from_meta(md):
    """
    Returns a list containing indices of the sync traces in the original array
    """
    typ = _get_type_from_meta(md)
    if typ != 'nidq':
        return []
    tr = md.get('snsMnMaXaDw')
    nsa = int(tr[-2])
    return list(range(int(sum(tr[0:2])), int(sum(tr[0:2])) + nsa))


def _get_nchannels_from_meta(md):
    typ = _get_type_from_meta(md)
    if typ == 'nidq':
        return int(np.round(np.sum(md.get('snsMnMaXaDw'))))
    elif typ in ['lf', 'ap']:
        return int(np.round(sum(md.get('snsApLfSy'))))


def _get_fs_from_meta(md):
    if md.get('typeThis') == 'imec':
        return md.get('imSampRate')
    else:
        return md.get('niSampRate')


def _get_type_from_meta(md):
    """
    Get neuropixel data type (ap, lf or nidq) from metadata
    """
    snsApLfSy = md.get('snsApLfSy', [-1, -1, -1])
    if snsApLfSy[0] == 0 and snsApLfSy[1] != 0:
        return 'lf'
    elif snsApLfSy[0] != 0 and snsApLfSy[1] == 0:
        return 'ap'
    elif snsApLfSy == [-1, -1, -1] and md.get('typeThis', None) == 'nidq':
        return 'nidq'


def _map_channels_from_meta(meta_data):
    """
    Interpret the meta data string to extract an array of channel positions along the shank

    :param meta_data: dictionary output from  spikeglx.read_meta_data
    :return: dictionary of arrays 'shank', 'col', 'row', 'flag', one value per active site
    """
    if 'snsShankMap' in meta_data.keys():
        chmap = re.findall(r'([0-9]*:[0-9]*:[0-9]*:[0-9]*)', meta_data['snsShankMap'])
        # for digital nidq types, the key exists but does not contain any information
        if not chmap:
            return {'shank': None, 'col': None, 'row': None, 'flag': None}
        # shank#, col#, row#, drawflag
        # (nb: drawflag is one should be drawn and considered spatial average)
        chmap = np.array([np.float32(cm.split(':')) for cm in chmap])
        return {k: chmap[:, v] for (k, v) in {'shank': 0, 'col': 1, 'row': 2, 'flag': 3}.items()}


def _conversion_sample2v_from_meta(meta_data):
    """
    Interpret the meta data to extract an array of conversion factors for each channel
    so the output data is in Volts
    Conversion factor is: int2volt / channelGain
    For Lf/Ap interpret the gain string from metadata
    For Nidq, repmat the gains from the trace counts in `snsMnMaXaDw`

    :param meta_data: dictionary output from  spikeglx.read_meta_data
    :return: numpy array with one gain value per channel
    """

    def int2volts(md):
        """ :return: Conversion scalar to Volts. Needs to be combined with channel gains """
        if md.get('typeThis', None) == 'imec':
            return md.get('imAiRangeMax') / 512
        else:
            return md.get('niAiRangeMax') / 32768

    int2volt = int2volts(meta_data)
    # interprets the gain value from the metadata header:
    if 'imroTbl' in meta_data.keys():  # binary from the probes: ap or lf
        sy_gain = np.ones(int(meta_data['snsApLfSy'][-1]), dtype=np.float32)
        # imroTbl has 384 entries regardless of no of channels saved, so need to index by n_ch
        n_chn = _get_nchannels_from_meta(meta_data) - 1
        # the sync traces are not included in the gain values, so are included for broadcast ops
        gain = re.findall(r'([0-9]* [0-9]* [0-9]* [0-9]* [0-9]*)', meta_data['imroTbl'])[:n_chn]
        out = {'lf': np.hstack((np.array([1 / np.float32(g.split(' ')[-1]) for g in gain]) *
                                int2volt, sy_gain)),
               'ap': np.hstack((np.array([1 / np.float32(g.split(' ')[-2]) for g in gain]) *
                                int2volt, sy_gain))}
    elif 'niMNGain' in meta_data.keys():  # binary from nidq
        gain = np.r_[
            np.ones(int(meta_data['snsMnMaXaDw'][0],)) / meta_data['niMNGain'] * int2volt,
            np.ones(int(meta_data['snsMnMaXaDw'][1],)) / meta_data['niMAGain'] * int2volt,
            np.ones(int(meta_data['snsMnMaXaDw'][2], )) * int2volt,  # no gain for analog sync
            np.ones(int(np.sum(meta_data['snsMnMaXaDw'][3]),))]  # no unit for digital sync
        out = {'nidq': gain}
    return out


def split_sync(sync_tr):
    """
    The synchronization channels are stored as single bits, this will split the int16 original
    channel into 16 single bits channels

    :param sync_tr: numpy vector: samples of synchronisation trace
    :return: int8 numpy array of 16 channels, 1 column per sync trace
    """
    sync_tr = np.int16(np.copy(sync_tr))
    out = np.unpackbits(sync_tr.view(np.uint8)).reshape(sync_tr.size, 16)
    out = np.flip(np.roll(out, 8, axis=1), axis=1)
    return np.int8(out)


def get_neuropixel_version_from_folder(session_path):
    ephys_files = glob_ephys_files(session_path)
    return get_neuropixel_version_from_files(ephys_files)


def get_neuropixel_version_from_files(ephys_files):
    if any([ef.get('nidq') for ef in ephys_files]):
        return '3B'
    else:
        return '3A'


def glob_ephys_files(session_path, suffix='.meta', recursive=True, bin_exists=True):
    """
    From an arbitrary folder (usually session folder) gets the ap and lf files and labels
    Associated to the subfolders where they are
    the expected folder tree is:
    ├── 3A
    │        ├── imec0
    │                ├── sync_testing_g0_t0.imec0.ap.bin
    │        │        └── sync_testing_g0_t0.imec0.lf.bin
    │        └── imec1
    │            ├── sync_testing_g0_t0.imec1.ap.bin
    │            └── sync_testing_g0_t0.imec1.lf.bin
    └── 3B
        ├── sync_testing_g0_t0.nidq.bin
        ├── imec0
        │        ├── sync_testing_g0_t0.imec0.ap.bin
        │        └── sync_testing_g0_t0.imec0.lf.bin
        └── imec1
            ├── sync_testing_g0_t0.imec1.ap.bin
            └── sync_testing_g0_t0.imec1.lf.bin

    :param bin_exists:
    :param suffix:
    :param recursive:
    :param session_path: folder, string or pathlib.Path
    :param glob_pattern: pattern to look recursively for (defaults to '*.ap.*bin)
    :returns: a list of dictionaries with keys 'ap': apfile, 'lf': lffile and 'label'
    """
    def get_label(raw_ephys_apfile):
        if raw_ephys_apfile.parts[-2] != 'raw_ephys_data':
            return raw_ephys_apfile.parts[-2]
        else:
            return ''

    recurse = '**/' if recursive else ''
    ephys_files = []
    for raw_ephys_file in Path(session_path).glob(f'{recurse}*.ap{suffix}'):
        raw_ephys_apfile = next(raw_ephys_file.parent.glob(raw_ephys_file.stem + '.*bin'), None)
        if not raw_ephys_apfile and bin_exists:
            continue
        elif not bin_exists:
            raw_ephys_apfile = raw_ephys_file.with_suffix('.bin')
        # first get the ap file
        ephys_files.extend([Bunch({'label': None, 'ap': None, 'lf': None, 'path': None})])
        ephys_files[-1].ap = raw_ephys_apfile
        # then get the corresponding lf file if it exists
        lf_file = raw_ephys_apfile.parent / raw_ephys_apfile.name.replace('.ap.', '.lf.')
        ephys_files[-1].lf = next(lf_file.parent.glob(lf_file.stem + '.*bin'), None)
        # finally, the label is the current directory except if it is bare in raw_ephys_data
        ephys_files[-1].label = get_label(raw_ephys_apfile)
        ephys_files[-1].path = raw_ephys_apfile.parent
    # for 3b probes, need also to get the nidq dataset type
    for raw_ephys_file in Path(session_path).rglob(f'{recurse}*.nidq{suffix}'):
        raw_ephys_nidqfile = next(raw_ephys_file.parent.glob(raw_ephys_file.stem + '.*bin'), None)
        if not bin_exists:
            raw_ephys_nidqfile = raw_ephys_file.with_suffix('.bin')
        ephys_files.extend([Bunch({'label': get_label(raw_ephys_file),
                                   'nidq': raw_ephys_nidqfile,
                                   'path': raw_ephys_file.parent})])
    return ephys_files


def _mock_spikeglx_file(mock_bin_file, meta_file, ns, nc, sync_depth,
                        random=False, int2volts=0.6 / 32768, corrupt=False):
    """
    For testing purposes, create a binary file with sync pulses to test reading and extraction
    """
    meta_file = Path(meta_file)
    mock_path_bin = Path(mock_bin_file)
    mock_path_meta = mock_path_bin.with_suffix('.meta')
    md = read_meta_data(meta_file)
    assert meta_file != mock_path_meta
    fs = _get_fs_from_meta(md)
    fid_source = open(meta_file)
    fid_target = open(mock_path_meta, 'w+')
    line = fid_source.readline()
    while line:
        line = fid_source.readline()
        if line.startswith('fileSizeBytes'):
            line = f'fileSizeBytes={ns * nc * 2}\n'
        if line.startswith('fileTimeSecs'):
            if corrupt:
                line = f'fileTimeSecs={ns / fs + 1.8324}\n'
            else:
                line = f'fileTimeSecs={ns / fs}\n'
        fid_target.write(line)
    fid_source.close()
    fid_target.close()
    if random:
        D = np.random.randint(-32767, 32767, size=(ns, nc), dtype=np.int16)
    else:  # each channel as an int of chn + 1
        D = np.tile(np.int16((np.arange(nc) + 1) / int2volts), (ns, 1))
        D[0:16, :] = 0
    # the last channel is the sync that we fill with
    sync = np.int16(2 ** np.float32(np.arange(-1, sync_depth)))
    D[:, -1] = 0
    D[:sync.size, -1] = sync
    with open(mock_path_bin, 'w+') as fid:
        D.tofile(fid)
    return {'bin_file': mock_path_bin, 'ns': ns, 'nc': nc, 'sync_depth': sync_depth, 'D': D}


def get_hardware_config(config_file):
    """
    Reads the neuropixel_wirings.json file containing sync mapping and parameters
    :param config_file: folder or json file
    :return: dictionary or None
    """
    config_file = Path(config_file)
    if config_file.is_dir():
        config_file = list(config_file.glob('*.wiring.json'))
        if config_file:
            config_file = config_file[0]
    if not config_file or not config_file.exists():
        return
    with open(config_file) as fid:
        par = json.loads(fid.read())
    return par


def _sync_map_from_hardware_config(hardware_config):
    """
    :param hardware_config: dictonary from json read of neuropixel_wirings.json
    :return: dictionary where key names refer to object and values to sync channel index
    """
    pin_out = neuropixel.SYNC_PIN_OUT[hardware_config['SYSTEM']]
    sync_map = {hardware_config['SYNC_WIRING_DIGITAL'][pin]: pin_out[pin]
                for pin in hardware_config['SYNC_WIRING_DIGITAL']
                if pin_out[pin] is not None}
    analog = hardware_config.get('SYNC_WIRING_ANALOG')
    if analog:
        sync_map.update({analog[pin]: int(pin[2:]) + 16 for pin in analog})
    return sync_map


def get_sync_map(folder_ephys):
    hc = get_hardware_config(folder_ephys)
    if not hc:
        _logger.warning(f"No channel map for {str(folder_ephys)}")
        return None
    else:
        return _sync_map_from_hardware_config(hc)



import scipy
import numpy as np
import ibllib.pipes.histology as histology
import ibllib.atlas as atlas
TIP_SIZE_UM = 200


def _cumulative_distance(xyz):
    return np.cumsum(np.r_[0, np.sqrt(np.sum(np.diff(xyz, axis=0) ** 2, axis=1))])


class EphysAlignment:

    def __init__(self, xyz_picks, chn_depths=None, track_prev=None,
                 feature_prev=None, brain_atlas=None):

        if not brain_atlas:
            self.brain_atlas = atlas.AllenAtlas(25)
        else:
            self.brain_atlas = brain_atlas

        self.xyz_track, self.track_extent = self.get_insertion_track(xyz_picks)
        self.chn_depths = chn_depths
        if np.any(track_prev):
            self.track_init = track_prev
            self.feature_init = feature_prev
        else:
            self.track_init = np.copy(self.track_extent)
            self.feature_init = np.copy(self.track_extent)

        self.sampling_trk = np.arange(self.track_extent[0],
                                      self.track_extent[-1] - 10 * 1e-6, 10 * 1e-6)
        self.xyz_samples = histology.interpolate_along_track(self.xyz_track,
                                                             self.sampling_trk -
                                                             self.sampling_trk[0])

        self.region, self.region_label, self.region_colour, self.region_id\
            = self.get_histology_regions(self.xyz_samples, self.sampling_trk, self.brain_atlas)

    def get_insertion_track(self, xyz_picks):
        """
        Extends probe trajectory from bottom of brain to upper bound of allen atlas
        :param xyz_picks: points defining probe trajectory in 3D space (xyz)
        :type xyz_picks: np.array((n, 3)) - n: no. of unique points
        :return xyz_track: points defining extended trajectory in 3D space (xyz)
        :type xyz_track: np.array((n+2, 3))
        :return track_extent: cumulative distance between two extremes of xyz_track (bottom of
        brain and top of atlas) offset by distance to probe tip
        :type track_extent: np.array((2))
        """
        # Use the first and last quarter of xyz_picks to estimate the trajectory beyond xyz_picks
        n_picks = np.max([4, round(xyz_picks.shape[0] / 4)])
        traj_entry = atlas.Trajectory.fit(xyz_picks[:n_picks, :])
        traj_exit = atlas.Trajectory.fit(xyz_picks[-1 * n_picks:, :])
        # Force the entry to be on the upper z lim of the atlas to account for cases where channels
        # may be located above the surface of the brain
        entry = (traj_entry.eval_z(self.brain_atlas.bc.zlim))[0, :]
        # The exit is just below the bottom surfacce of the brain
        exit = atlas.Insertion.get_brain_exit(traj_exit, self.brain_atlas)
        exit[2] = exit[2] - 200 / 1e6

        xyz_track = np.r_[exit[np.newaxis, :], xyz_picks, entry[np.newaxis, :]]
        # Sort so that most ventral coordinate is first
        xyz_track = xyz_track[np.argsort(xyz_track[:, 2]), :]

        # Compute distance to first electrode from bottom coordinate
        tip_distance = _cumulative_distance(xyz_track)[1] + TIP_SIZE_UM / 1e6
        track_length = _cumulative_distance(xyz_track)[-1]
        track_extent = np.array([0, track_length]) - tip_distance

        return xyz_track, track_extent

    def get_track_and_feature(self):
        """
        Return track, feature and xyz_track variables
        """
        return self.feature_init, self.track_init, self.xyz_track

    @staticmethod
    def feature2track(trk, feature, track):
        """
        Estimate new values of trk according to interpolated fit between feature and track space
        :param trk: points in track space to convert feature space
        :type trk: np.array
        :param feature: reference coordinates in feature space (ephys plots)
        :type feature: np.array((n_lines + 2)) n_lines: no. of user reference lines
        :param track: reference coordinates in track space (histology track)
        :type track: np.array((n_lines + 2))
        :return fcn(trk): interpolated values of trk
        :type fcn(trk): np.array
        """

        fcn = scipy.interpolate.interp1d(feature, track, fill_value="extrapolate")
        return fcn(trk)

    @staticmethod
    def track2feature(ft, feature, track):
        """
        Estimate new values of ft according to interpolated fit between track and feature space
        :param ft: points in feature space to convert track space
        :type ft: np.array
        :param feature: reference coordinates in feature space (ephys plots)
        :type feature: np.array((n_lines + 2)) n_lines: no. of user reference lines
        :param track: reference coordinates in track space (histology track)
        :type track: np.array((n_lines + 2))
        :return fcn(ft): interpolated values of ft
        :type fcn(ft): np.array
        """
        fcn = scipy.interpolate.interp1d(track, feature, fill_value="extrapolate")
        return fcn(ft)

    @staticmethod
    def feature2track_lin(trk, feature, track):
        """
        Estimate new values of trk according to linear fit between feature and track space, only
        implemented if no. of reference points >= 3
        :param trk: points in track space to convert feature space
        :type trk: np.array
        :param feature: reference coordinates in feature space (ephys plots)
        :type feature: np.array((n_lines + 2)) n_lines: no. of user reference lines
        :param track: reference coordinates in track space (histology track)
        :type track: np.array((n_lines + 2))
        :return fcn(trk): linear fit values of trk
        :type fcn(trk): np.array
        """
        if feature.size >= 5:
            fcn_lin = np.poly1d(np.polyfit(feature[1:-1], track[1:-1], 1))
            lin_fit = fcn_lin(trk)
        else:
            lin_fit = 0
        return lin_fit

    @staticmethod
    def adjust_extremes_uniform(feature, track):
        """
        Change the value of the first and last reference points (non user chosen points) such
        that coordinates outside user picked regions are left unchanged
        :param feature: reference coordinates in feature space (ephys plots)
        :type feature: np.array((n_lines + 2)) n_lines: no. of user reference lines
        :param track: reference coordinates in track space (histology track)
        :type track: np.array((n_lines + 2))
        :return track: reference coordinates in track space with first and last value adjusted
        :type track: np.array((n_lines + 2))
        """
        diff = np.diff(feature - track)
        track[0] -= diff[0]
        track[-1] += diff[-1]
        return track

    def adjust_extremes_linear(self, feature, track, extend_feature=1):
        """
        Change the value of the first and last reference points (non user chosen points) such
        that coordinates outside user picked regions have a linear fit applied
        :param feature: reference coordinates in feature space (ephys plots)
        :type feature: np.array((n_lines + 2)) n_lines: no. of user reference lines
        :param track: reference coordinates in track space (histology track)
        :type track: np.array((n_lines + 2))
        :param extend_feature: amount to extend extreme coordinates before applying linear fit
        :type extend_feature: float
        :return feature: reference coordinates in feature space with first and last value adjusted
        :type feature: np.array((n_lines + 2))
        :return track: reference coordinates in track space with first and last value adjusted
        :type track: np.array((n_lines + 2))
        """
        feature[0] = self.track_extent[0] - extend_feature
        feature[-1] = self.track_extent[-1] + extend_feature
        extend_track = self.feature2track_lin(feature[[0, -1]], feature, track)
        track[0] = extend_track[0]
        track[-1] = extend_track[-1]
        return feature, track

    def scale_histology_regions(self, feature, track):
        """
        Recompute locations of brain region boundaries using interpolated fit based on reference
        lines
        :param feature: reference coordinates in feature space (ephys plots)
        :type feature: np.array((n_lines + 2)) n_lines: no. of user reference lines
        :param track: reference coordinates in track space (histology track)
        :type track: np.array((n_lines + 2))
        :return region: new coordinates of histology boundaries after applying interpolation
        :type region: np.array((n_bound, 2)) n_bound: no. of histology boundaries
        :return region_label: new coordinates of histology labels positions after applying
                              interpolation
        :type region_label: np.array((n_bound)) of tuples (coordinate - float, label - str)
        """
        region_label = np.copy(self.region_label)
        region = self.track2feature(self.region, feature, track) * 1e6
        region_label[:, 0] = (self.track2feature(np.float64(region_label[:, 0]), feature,
                              track) * 1e6)
        return region, region_label

    @staticmethod
    def get_histology_regions(xyz_coords, depth_coords, brain_atlas=None):
        """
        Find all brain regions and their boundaries along the depth of probe or track
        :param xyz_coords: 3D coordinates of points along probe or track
        :type xyz_coords: np.array((n_points, 3)) n_points: no. of points
        :param depth_coords: depth along probe or track where each xyz_coord is located
        :type depth_coords: np.array((n_points))
        :return region: coordinates bounding each brain region
        :type region: np.array((n_bound, 2)) n_bound: no. of histology boundaries
        :return region_label: label for each brain region and coordinate of where to place label
        :type region_label: np.array((n_bound)) of tuples (coordinate - float, label - str)
        :return region_colour: allen atlas rgb colour for each brain region along track
        :type region_colour: np.array((n_bound, 3))
        :return region_id: allen atlas id for each brain region along track
        :type region_id: np.array((n_bound))
        """
        if not brain_atlas:
            brain_atlas = atlas.AllenAtlas(25)

        region_ids = brain_atlas.get_labels(xyz_coords)
        region_info = brain_atlas.regions.get(region_ids)
        boundaries = np.where(np.diff(region_info.id))[0]
        region = np.empty((boundaries.size + 1, 2))
        region_label = np.empty((boundaries.size + 1, 2), dtype=object)
        region_id = np.empty((boundaries.size + 1, 1), dtype=int)
        region_colour = np.empty((boundaries.size + 1, 3), dtype=int)
        for bound in np.arange(boundaries.size + 1):
            if bound == 0:
                _region = np.array([0, boundaries[bound]])
            elif bound == boundaries.size:
                _region = np.array([boundaries[bound - 1], region_info.id.size - 1])
            else:
                _region = np.array([boundaries[bound - 1], boundaries[bound]])
            _region_colour = region_info.rgb[_region[1]]
            _region_label = region_info.acronym[_region[1]]
            _region_id = region_info.id[_region[1]]
            _region = depth_coords[_region]
            _region_mean = np.mean(_region)
            region[bound, :] = _region
            region_colour[bound, :] = _region_colour
            region_id[bound, :] = _region_id
            region_label[bound, :] = (_region_mean, _region_label)

        return region, region_label, region_colour, region_id

    @staticmethod
    def get_nearest_boundary(xyz_coords, allen, extent=100, steps=8, parent=True,
                             brain_atlas=None):
        """
        Finds distance to closest neighbouring brain region along trajectory. For each point in
        xyz_coords computes the plane passing through point and perpendicular to trajectory and
        finds all brain regions that lie in that plane up to a given distance extent from specified
        point. Additionally, if requested, computes distance between the parents of regions.
        :param xyz_coords: 3D coordinates of points along probe or track
        :type xyz_coords: np.array((n_points, 3)) n_points: no. of points
        :param allen: dataframe containing allen info. Loaded from allen_structure_tree in
        ibllib/atlas
        :type allen: pandas Dataframe
        :param extent: extent of plane in each direction from origin in (um)
        :type extent: float
        :param steps: no. of steps to discretise plane into
        :type steps: int
        :param parent: Whether to also compute nearest distance between parents of regions
        :type parent: bool
        :return nearest_bound: dict containing results
        :type nearest_bound: dict
        """
        if not brain_atlas:
            brain_atlas = atlas.AllenAtlas(25)

        vector = atlas.Insertion.from_track(xyz_coords, brain_atlas=brain_atlas).trajectory.vector
        nearest_bound = dict()
        nearest_bound['dist'] = np.zeros((xyz_coords.shape[0]))
        nearest_bound['id'] = np.zeros((xyz_coords.shape[0]))
        # nearest_bound['adj_id'] = np.zeros((xyz_coords.shape[0]))
        nearest_bound['col'] = []

        if parent:
            nearest_bound['parent_dist'] = np.zeros((xyz_coords.shape[0]))
            nearest_bound['parent_id'] = np.zeros((xyz_coords.shape[0]))
            # nearest_bound['parent_adj_id'] = np.zeros((xyz_coords.shape[0]))
            nearest_bound['parent_col'] = []

        for iP, point in enumerate(xyz_coords):
            d = np.dot(vector, point)
            x_vals = np.r_[np.linspace(point[0] - extent / 1e6, point[0] + extent / 1e6, steps),
                           point[0]]
            y_vals = np.r_[np.linspace(point[1] - extent / 1e6, point[1] + extent / 1e6, steps),
                           point[1]]

            X, Y = np.meshgrid(x_vals, y_vals)
            Z = (d - vector[0] * X - vector[1] * Y) / vector[2]
            XYZ = np.c_[np.reshape(X, X.size), np.reshape(Y, Y.size), np.reshape(Z, Z.size)]
            dist = np.sqrt(np.sum((XYZ - point) ** 2, axis=1))

            try:
                brain_id = brain_atlas.regions.get(brain_atlas.get_labels(XYZ))['id']
            except Exception as err:
                print(err)
                continue

            dist_sorted = np.argsort(dist)
            brain_id_sorted = brain_id[dist_sorted]
            nearest_bound['id'][iP] = brain_id_sorted[0]
            nearest_bound['col'].append(allen['color_hex_triplet'][np.where(allen['id'] ==
                                                                   brain_id_sorted[0])[0][0]])
            bound_idx = np.where(brain_id_sorted != brain_id_sorted[0])[0]
            if np.any(bound_idx):
                nearest_bound['dist'][iP] = dist[dist_sorted[bound_idx[0]]] * 1e6
                # nearest_bound['adj_id'][iP] = brain_id_sorted[bound_idx[0]]
            else:
                nearest_bound['dist'][iP] = np.max(dist) * 1e6
                # nearest_bound['adj_id'][iP] = brain_id_sorted[0]

            if parent:
                # Now compute for the parents
                brain_parent = np.array([allen['parent_structure_id'][np.where(allen['id'] == br)
                                        [0][0]] for br in brain_id_sorted])
                brain_parent[np.isnan(brain_parent)] = 0

                nearest_bound['parent_id'][iP] = brain_parent[0]
                nearest_bound['parent_col'].append(allen['color_hex_triplet']
                                                   [np.where(allen['id'] ==
                                                             brain_parent[0])[0][0]])

                parent_idx = np.where(brain_parent != brain_parent[0])[0]
                if np.any(parent_idx):
                    nearest_bound['parent_dist'][iP] = dist[dist_sorted[parent_idx[0]]] * 1e6
                    # nearest_bound['parent_adj_id'][iP] = brain_parent[parent_idx[0]]
                else:
                    nearest_bound['parent_dist'][iP] = np.max(dist) * 1e6
                    # nearest_bound['parent_adj_id'][iP] = brain_parent[0]

        return nearest_bound

    @staticmethod
    def arrange_into_regions(depth_coords, region_ids, distance, region_colours):
        """
        Arrange output from get_nearest_boundary into a form that can be plot using pyqtgraph or
        matplotlib
        :param depth_coords: depth along probe or track where each point is located
        :type depth_coords: np.array((n_points))
        :param region_ids: brain region id at each depth along probe
        :type regions_ids: np.array((n_points))
        :param distance: distance to nearest boundary in plane at each point
        :type distance: np.array((n_points))
        :param region_colours: allen atlas hex colour for each region id
        :type region_colours: list of strings len(n_points)
        :return all_x: dist values for each region along probe track
        :type all_x: list of np.array
        :return all_y: depth values for each region along probe track
        :type all_y: list of np.array
        :return all_colour: colour assigned to each region along probe track
        :type all_colour: list of str
        """

        boundaries = np.where(np.diff(region_ids))[0]
        bound = np.r_[0, boundaries + 1, region_ids.shape[0]]
        all_y = []
        all_x = []
        all_colour = []
        for iB in np.arange(len(bound) - 1):
            y = depth_coords[bound[iB]:(bound[iB + 1])]
            y = np.r_[y[0], y, y[-1]]
            x = distance[bound[iB]:(bound[iB + 1])]
            x = np.r_[0, x, 0]
            all_y.append(y)
            all_x.append(x)
            col = region_colours[bound[iB]]
            if type(col) != str:
                col = '#FFFFFF'
            else:
                col = '#' + col
            all_colour.append(col)

        return all_x, all_y, all_colour

    def get_scale_factor(self, region):
        """
        Find how much each brain region has been scaled following interpolation
        :param region: scaled histology boundaries
        :type region: np.array((n_bound, 2)) n_bound: no. of histology boundaries
        :return scaled_region: regions that have unique scaling applied
        :type scaled_region: np.array((n_scale, 2)) n_scale: no. of uniquely scaled regions
        :return scale_factor: scale factor applied to each scaled region
        :type scale_factor: np.array((n_scale))
        """
        scale = []
        for iR, (reg, reg_orig) in enumerate(zip(region, self.region * 1e6)):
            scale = np.r_[scale, (reg[1] - reg[0]) / (reg_orig[1] - reg_orig[0])]
        boundaries = np.where(np.diff(np.around(scale, 3)))[0]
        if boundaries.size == 0:
            scaled_region = np.array([[region[0][0], region[-1][1]]])
            scale_factor = np.unique(scale)
        else:
            scaled_region = np.empty((boundaries.size + 1, 2))
            scale_factor = []
            for bound in np.arange(boundaries.size + 1):
                if bound == 0:
                    _scaled_region = np.array([region[0][0],
                                              region[boundaries[bound]][1]])
                    _scale_factor = scale[0]
                elif bound == boundaries.size:
                    _scaled_region = np.array([region[boundaries[bound - 1]][1],
                                              region[-1][1]])
                    _scale_factor = scale[-1]
                else:
                    _scaled_region = np.array([region[boundaries[bound - 1]][1],
                                              region[boundaries[bound]][1]])
                    _scale_factor = scale[boundaries[bound]]
                scaled_region[bound, :] = _scaled_region
                scale_factor = np.r_[scale_factor, _scale_factor]
        return scaled_region, scale_factor

    def get_channel_locations(self, feature, track, depths=None):
        """
        Gets 3d coordinates from a depth along the electrophysiology feature. 2 steps
        1) interpolate from the electrophys features depths space to the probe depth space
        2) interpolate from the probe depth space to the true 3D coordinates
        if depths is not provided, defaults to channels local coordinates depths
        """
        if depths is None:
            depths = self.chn_depths / 1e6
        # nb using scipy here so we can change to cubic spline if needed
        channel_depths_track = self.feature2track(depths, feature, track) - self.track_extent[0]
        xyz_channels = histology.interpolate_along_track(self.xyz_track, channel_depths_track)
        return xyz_channels

    def get_brain_locations(self, xyz_channels):
        """
        Finds the brain regions from 3D coordinates of electrode locations
        :param xyz_channels: 3D coordinates of electrodes on probe
        :type xyz_channels: np.array((n_elec, 3)) n_elec: no. of electrodes (384)
        :return brain_regions: brain region object for each electrode
        :type dict
        """
        brain_regions = self.brain_atlas.regions.get(self.brain_atlas.get_labels(xyz_channels))
        return brain_regions

    def get_perp_vector(self, feature, track):
        """
        Finds the perpendicular vector along the trajectory at the depth of reference lines
        :param feature: reference coordinates in feature space (ephys plots)
        :type feature: np.array((n_lines + 2)) n_lines: no. of user reference lines
        :param track: reference coordinates in track space (histology track)
        :type track: np.array((n_line+2))
        :return slice_lines: coordinates of perpendicular lines
        :type slice_lines: np.array((n_lines, 2))
        """

        slice_lines = []
        for line in feature[1:-1]:
            depths = np.array([line, line + 10 / 1e6])
            xyz = self.get_channel_locations(feature, track, depths)

            extent = 500e-6
            vector = np.diff(xyz, axis=0)[0]
            point = xyz[0, :]
            vector_perp = np.array([1, 0, -1 * vector[0] / vector[2]])
            xyz_per = np.r_[[point + (-1 * extent * vector_perp)],
                            [point + (extent * vector_perp)]]
            slice_lines.append(xyz_per)

        return slice_lines



# library of small functions
import datetime
import numpy as np


def isostr2date(isostr):
    """
    Convert strings representing dates into datetime.datetime objects aimed ad Django REST API
    ISO 8601: '2018-05-22T14:35:22.99585' or '2018-05-22T14:35:22'

    :param isostr: a string, list of strings or panda Series / numpy arrays containing strings
    :return: a scalar, list of
    """
    # NB this is intended for scalars or small list. See the ciso8601 pypi module instead for
    # a performance implementation
    if not isinstance(isostr, str):
        return [isostr2date(el) for el in isostr]

    format = '%Y-%m-%dT%H:%M:%S'
    if '.' in isostr:
        format += '.%f'
    if '+' in isostr:
        format += '.%f'
    return datetime.datetime.strptime(isostr, format)


def date2isostr(adate):
    # NB this is intended for scalars or small list. See the ciso8601 pypi module instead for
    # a performance implementation
    if type(adate) is datetime.date:
        adate = datetime.datetime.fromordinal(adate.toordinal())
    return datetime.datetime.isoformat(adate)


def format_date_range(date_range):
    if all([isinstance(d, str) for d in date_range]):
        date_range = [datetime.datetime.strptime(d, '%Y-%m-%d') for d in date_range]
    elif not all([isinstance(d, datetime.date) for d in date_range]):
        raise ValueError('Date range doesn''t have proper format: list of 2 strings "yyyy-mm-dd" ')
    # the django filter is implemented in datetime and assumes the beginning of the day (last day
    # is excluded by default
    date_range = [d.strftime('%Y-%m-%d') for d in date_range]
    return date_range


def convert_pgts(time):
    """Convert PointGray cameras timestamps to seconds.
    Use convert then uncycle"""
    # offset = time & 0xFFF
    cycle1 = (time >> 12) & 0x1FFF
    cycle2 = (time >> 25) & 0x7F
    seconds = cycle2 + cycle1 / 8000.
    return seconds


def uncycle_pgts(time):
    """Unwrap the converted seconds of a PointGray camera timestamp series."""
    cycles = np.insert(np.diff(time) < 0, 0, False)
    cycleindex = np.cumsum(cycles)
    return time + cycleindex * 128



import unittest
import tempfile
from pathlib import Path
import json

import numpy as np

from ibllib.misc import version
from oneibl import one, registration

one = one.ONE(base_url='https://test.alyx.internationalbrainlab.org', username='test_user',
              password='TapetesBloc18')
SUBJECT = 'clns0730'
USER = 'test_user'

# one = one.ONE(base_url='http://localhost:8000')
# SUBJECT = 'CSP013'
# USER = 'olivier'

md5_0 = 'add2ab27dbf8428f8140-0870d5080c7f'
r = {'created_by': 'olivier',
     'path': f'{SUBJECT}/2018-08-24/002',
     'filenames': ["raw_behavior_data/_iblrig_encoderTrialInfo.raw.ssv"],
     'hashes': [md5_0],
     'filesizes': [1234],
     'versions': [version.ibllib()]}

MOCK_SESSION_SETTINGS = {
    'SESSION_DATE': '2018-04-01',
    'SESSION_DATETIME': '2018-04-01T12:48:26.795526',
    'PYBPOD_CREATOR': [USER,
                       'f092c2d5-c98a-45a1-be7c-df05f129a93c',
                       'local'],
    'SESSION_NUMBER': '002',
    'SUBJECT_NAME': SUBJECT,
    'PYBPOD_BOARD': '_iblrig_mainenlab_behavior_1',
    'PYBPOD_PROTOCOL': '_iblrig_tasks_ephysChoiceWorld',
    'IBLRIG_VERSION_TAG': '5.4.1',
    'SUBJECT_WEIGHT': 22,
}

MOCK_SESSION_DICT = {
    'subject': SUBJECT,
    'start_time': '2018-04-01T12:48:26.795526',
    'number': 2,
    'users': [USER]
}


class TestRegistrationEndpoint(unittest.TestCase):

    def test_task_names_extractors(self):
        """
        This is to test against regressions
        """
        task_out = [
            ('_iblrig_tasks_biasedChoiceWorld3.7.0', 'Behavior training/tasks'),
            ('_iblrig_tasks_biasedScanningChoiceWorld5.2.3', 'Behavior training/tasks'),
            ('_iblrig_tasks_trainingChoiceWorld3.6.0', 'Behavior training/tasks'),
            ('_iblrig_tasks_ephysChoiceWorld5.1.3', 'Ephys recording with acute probe(s)'),
            ('_iblrig_calibration_frame2TTL4.1.3', None),
            ('_iblrig_tasks_habituationChoiceWorld3.6.0', 'Behavior training/tasks'),
            ('_iblrig_tasks_scanningOptoChoiceWorld5.0.0', None),
            ('_iblrig_tasks_RewardChoiceWorld4.1.3', None),
            ('_iblrig_calibration_screen4.1.3', None),
            ('_iblrig_tasks_ephys_certification4.1.3', 'Ephys recording with acute probe(s)'),
        ]
        for to in task_out:
            out = registration._alyx_procedure_from_task(to[0])
            self.assertEqual(out, to[1])


class TestRegistration(unittest.TestCase):

    def setUp(self) -> None:
        # makes sure tests start without session created
        eid = one.search(subjects=SUBJECT, date_range='2018-04-01')
        for ei in eid:
            one.alyx.rest('sessions', 'delete', id=ei)
        self.td = tempfile.TemporaryDirectory()
        self.session_path = Path(self.td.name).joinpath(SUBJECT, '2018-04-01', '002')
        self.alf_path = self.session_path.joinpath('alf')
        self.alf_path.mkdir(parents=True)
        np.save(self.alf_path.joinpath('spikes.times.npy'), np.random.random(500))
        np.save(self.alf_path.joinpath('spikes.amps.npy'), np.random.random(500))

    def test_registration_datasets(self):
        # registers a single file
        ses = one.alyx.rest('sessions', 'create', data=MOCK_SESSION_DICT)
        st_file = self.alf_path.joinpath('spikes.times.npy')
        registration.register_dataset(file_list=st_file, one=one)
        dsets = one.alyx.rest('datasets', 'list', session=ses['url'][-36:])
        self.assertTrue(len(dsets) == 1)
        # registers a list of files
        flist = list(self.alf_path.glob('*.npy'))
        r = registration.register_dataset(file_list=flist, one=one)
        dsets = one.alyx.rest('datasets', 'list', session=ses['url'][-36:])
        self.assertTrue(len(dsets) == 2)
        # simulate all the datasets exists, re-register and asserts that exists is set to True
        # as the files haven't changed
        frs = one.alyx.rest('files', 'list', django=f"dataset__session,{ses['url'][-36:]}")
        for fr in frs:
            one.alyx.rest('files', 'partial_update', id=fr['url'][-36:], data={'exists': True})
        r = registration.register_dataset(file_list=flist, one=one)
        self.assertTrue(all([all([fr['exists'] for fr in rr['file_records']]) for rr in r]))
        # now that files have changed, makes sure the exists flags are set to False
        np.save(self.alf_path.joinpath('spikes.times.npy'), np.random.random(500))
        np.save(self.alf_path.joinpath('spikes.amps.npy'), np.random.random(500))
        r = registration.register_dataset(file_list=flist, one=one)
        self.assertTrue(all([all([not(fr['exists']) for fr in rr['file_records']]) for rr in r]))

    def test_registration_session(self):
        behavior_path = self.session_path.joinpath('raw_behavior_data')
        behavior_path.mkdir()
        settings_file = behavior_path.joinpath('_iblrig_taskSettings.raw.json')
        with open(settings_file, 'w') as fid:
            json.dump(MOCK_SESSION_SETTINGS, fid)
        rc = registration.RegistrationClient(one=one)
        rc.register_session(self.session_path)
        eid = one.search(subjects=SUBJECT, date_range=['2018-04-01', '2018-04-01'])[0]
        datasets = one.alyx.rest('datasets', 'list', session=eid)
        for ds in datasets:
            self.assertTrue(ds['hash'] is not None)
            self.assertTrue(ds['file_size'] is not None)
            self.assertTrue(ds['version'] == version.ibllib())
        # checks the procedure of the session
        ses_info = one.alyx.rest('sessions', 'read', id=eid)
        self.assertTrue(ses_info['procedures'] == ['Ephys recording with acute probe(s)'])
        one.alyx.rest('sessions', 'delete', id=eid)
        # re-register the session as behaviour this time
        MOCK_SESSION_SETTINGS['PYBPOD_PROTOCOL'] = '_iblrig_tasks_trainingChoiceWorld6.3.1'
        with open(settings_file, 'w') as fid:
            json.dump(MOCK_SESSION_SETTINGS, fid)
        rc.register_session(self.session_path)
        eid = one.search(subjects=SUBJECT, date_range=['2018-04-01', '2018-04-01'])[0]
        ses_info = one.alyx.rest('sessions', 'read', id=eid)
        self.assertTrue(ses_info['procedures'] == ['Behavior training/tasks'])

    def tearDown(self) -> None:
        self.td.cleanup()


if __name__ == '__main__':
    unittest.main(exit=False)



""" by Luigi Acerbi, Shan Shen and Anne Urai
International Brain Laboratory, 2019

#TODO: Anne
"""

import numpy as np


class DataSet:
    """ Object containing data that can be preprocessed.
    Inputs: dataframe # pandas
    Methods: preprocess
    """


class TrialData:

    def __init__(self, data, meta_data=dict()):
        """data should contain the following columns:
        stimulus_side       : values to be -1 or 1
        stimulus_strength   : non-negative contrast
        choice              : -1 and 1, nan for missed trials
        rewarded            : 0 and 1, including 0 contrast trials
        ---
        optional columns:
        correct             : 1 for correct, 0 for incorrect, nan for 0 contrast or missed trials
        reaction_time       : time diff of response time - stim on time
        prob_left_block     : probability (in block structure) of stimulus_side == -1
        trial_id            :
        session_id          :
        """

        self.meta_data = meta_data
        self.trials_df = self.fill_dataframe(data)

    def fill_dataframe(self, data):
        data['signed_stimulus'] = data['stimulus_strength'] * data['stimulus_side']
        data['choice_right'] = data['choice'].replace([-1, 0, 1], [0, np.nan, 1])
        return data



from oneibl.one import ONE
import unittest


class Tests_REST(unittest.TestCase):

    def setUp(self):
        # Init connection to the database
        one = ONE(username='test_user', password='TapetesBloc18',
                  base_url='https://test.alyx.internationalbrainlab.org')
        self.one = one

    def test_water_restriction(self):
        """
        Examples of how to list all water restrictions and water-restriction for a given
        subject.
        """
        one = self.one
        # get all the water restrictions from start
        all_wr = one.alyx.rest('water-restriction', 'list')
        # 2 different ways to  get water restrictions for one subject
        wr_sub1 = one.alyx.rest('water-restriction', 'list', '?subject=algernon')
        wr_sub2 = one.alyx.rest('water-restriction', 'list', subject='algernon')  # recommended

        # enforce test logic
        self.assertTrue(set({'end_time', 'reference_weight', 'start_time', 'subject',
                             'water_type'}) >= set(all_wr[0].keys()))
        self.assertEqual(wr_sub1, wr_sub2)



"""
The psychofit toolbox contains tools to fit two-alternative psychometric
data. The fitting is done using maximal likelihood estimation: one
assumes that the responses of the subject are given by a binomial
distribution whose mean is given by the psychometric function.
The data can be expressed in fraction correct (from .5 to 1) or in
fraction of one specific choice (from 0 to 1). To fit them you can use
these functions:
  weibull50          - Weibull function from 0.5 to 1, with lapse rate
  weibull            - Weibull function from 0 to 1, with lapse rate
  erf_psycho         - erf function from 0 to 1, with lapse rate
  erf_psycho_2gammas - erf function from 0 to 1, with two lapse rates
Functions in the toolbox are:
  mle_fit_psycho     - Maximumum likelihood fit of psychometric function
  neg_likelihood     - Negative likelihood of a psychometric function
For more info, see:
  Examples           - Examples of use of psychofit toolbox
Matteo Carandini, 2000-2015
"""

import functools
import numpy as np
import scipy.optimize
from scipy.special import erf


def mle_fit_psycho(data, P_model='weibull', parstart=None, parmin=None, parmax=None, nfits=5):
    """
    Maximumum likelihood fit of psychometric function.
    Args:
        data: 3 x n matrix where first row corresponds to stim levels,
              the second to number of trials for each stim level (int),
              the third to proportion correct / proportion rightward (float between 0 and 1)
        P_model: The psychometric function. Possibilities include 'weibull'
                 (DEFAULT), 'weibull50', 'erf_psycho' and 'erf_psycho_2gammas'
        parstart: Non-zero starting parameters, used to try to avoid local minima.
                  The parameters are [threshold, slope, gamma], or if using the
                  'erf_psycho_2gammas' model append a second gamma value.
                  Recommended to use a value > 1. If None, some reasonable defaults are used.
        parmin: Minimum parameter values.  If None, some reasonable defaults are used
        parmax: Maximum parameter values.  If None, some reasonable defaults are used
        nfits: The number of fits
    Returns:
        pars: The parameters from the best of the fits
        L: The likelihood of the best fit
    Raises:
        TypeError: data must be a list or numpy array
        ValueError: data must be m by 3 matrix
    Examples:
        Below we fit a Weibull function to some data:
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> cc = np.array([-8., -6., -4., -2.,  0.,  2.,  4.,  6.,  8.]) # contrasts
        >>> nn = np.full((9,), 10) # number of trials at each contrast
        >>> pp = np.array([5., 8., 20., 41., 54., 59., 79., 92., 96])/100 # proportion "rightward"
        >>> pars, L = mle_fit_psycho(np.vstack((cc, nn, pp)), 'erf_psycho')
        >>> plt.plot(cc, pp, 'bo', mfc='b')
        >>> plt.plot(np.arange(-8, 8, 0.1), erf_psycho(pars, np.arange(-8, 8, 0.1)), '-b')
    Information:
        1999-11 FH wrote it
        2000-01 MC cleaned it up
        2000-04 MC took care of the 50% case
        2009-12 MC replaced fmins with fminsearch
        2010-02 MC, AZ added nfits
        2013-02 MC+MD fixed bug with dealing with NaNs
        2018-08 MW ported to Python
    """
    # Input validation
    if isinstance(data, (list, tuple)):
        data = np.array(data)
    elif not isinstance(data, np.ndarray):
        raise TypeError('data must be a list or numpy array')

    if data.shape[0] != 3:
        raise ValueError('data must be m by 3 matrix')

    rep = lambda x: (x, x) if P_model.endswith('2gammas') else (x,) # noqa
    if parstart is None:
        parstart = np.array([np.mean(data[0, :]), 3., *rep(.05)])
    if parmin is None:
        parmin = np.array([np.min(data[0, :]), 0., *rep(0.)])
    if parmax is None:
        parmax = np.array([np.max(data[0, :]), 10., *rep(.4)])

    # find the good values in pp (conditions that were effectively run)
    ii = np.isfinite(data[2, :])

    likelihoods = np.zeros(nfits,)
    pars = np.empty((nfits, parstart.size))

    f = functools.partial(neg_likelihood, data=data[:, ii],
                          P_model=P_model, parmin=parmin, parmax=parmax)
    for ifit in range(nfits):
        pars[ifit, :] = scipy.optimize.fmin(f, parstart, disp=False)
        parstart = parmin + np.random.rand(parmin.size) * (parmax - parmin)
        likelihoods[ifit] = -neg_likelihood(pars[ifit, :], data[:, ii], P_model, parmin, parmax)

    # the values to be output
    L = likelihoods.max()
    iBestFit = likelihoods.argmax()
    return pars[iBestFit, :], L


def neg_likelihood(pars, data, P_model='weibull', parmin=None, parmax=None):
    """
    Negative likelihood of a psychometric function.
    Args:
        pars: Model parameters [threshold, slope, gamma], or if
              using the 'erf_psycho_2gammas' model append a second gamma value.
        data: 3 x n matrix where first row corresponds to stim levels,
              the second to number of trials for each stim level (int),
              the third to proportion correct / proportion rightward (float between 0 and 1)
        P_model: The psychometric function. Possibilities include 'weibull'
                 (DEFAULT), 'weibull50', 'erf_psycho' and 'erf_psycho_2gammas'
        parmin: Minimum bound for parameters.  If None, some reasonable defaults are used
        parmax: Maximum bound for parameters.  If None, some reasonable defaults are used
    Returns:
        ll: The likelihood of the parameters.  The equation is:
            - sum(nn.*(pp.*log10(P_model)+(1-pp).*log10(1-P_model)))
            See the the appendix of Watson, A.B. (1979). Probability
            summation over time. Vision Res 19, 515-522.
    Raises:
        ValueError: invalid model, options are "weibull",
                    "weibull50", "erf_psycho" and "erf_psycho_2gammas"
        TypeError: data must be a list or numpy array
        ValueError data must be m by 3 matrix
    Information:
        1999-11 FH wrote it
        2000-01 MC cleaned it up
        2000-07 MC made it indep of Weibull and added parmin and parmax
        2018-08 MW ported to Python
    """
    # Validate input
    if isinstance(data, (list, tuple)):
        data = np.array(data)
    elif not isinstance(data, np.ndarray):
        raise TypeError('data must be a list or numpy array')

    if parmin is None:
        parmin = np.array([.005, 0., 0.])
    if parmax is None:
        parmax = np.array([.5, 10., .25])

    if data.shape[0] == 3:
        xx = data[0, :]
        nn = data[1, :]
        pp = data[2, :]
    else:
        raise ValueError('data must be m by 3 matrix')

    # here is where you effectively put the constraints.
    if (any(pars < parmin)) or (any(pars > parmax)):
        ll = 10000000
        return ll

    dispatcher = {
        'weibull': weibull,
        'weibull50': weibull50,
        'erf_psycho': erf_psycho,
        'erf_psycho_2gammas': erf_psycho_2gammas
    }
    try:
        probs = dispatcher[P_model](pars, xx)
    except KeyError:
        raise ValueError('invalid model, options are "weibull", ' +
                         '"weibull50", "erf_psycho" and "erf_psycho_2gammas"')

    assert (max(probs) <= 1) or (min(probs) >= 0), 'At least one of the probabilities is not ' \
                                                   'between 0 and 1'

    probs[probs == 0] = np.finfo(float).eps
    probs[probs == 1] = 1 - np.finfo(float).eps

    ll = - sum(nn * (pp * np.log(probs) + (1 - pp) * np.log(1 - probs)))
    return ll


def weibull(pars, xx):
    """
    Weibull function from 0 to 1, with lapse rate.
    Args:
        pars: Model parameters [alpha, beta, gamma].
        xx: vector of stim levels.
    Returns:
        A vector of length xx
    Raises:
        ValueError: pars must be of length 3
        TypeError: pars must be list-like or numpy array
    Information:
        1999-11 FH wrote it
        2000-01 MC cleaned it up
        2018-08 MW ported to Python
    """
    # Validate input
    if not isinstance(pars, (list, tuple, np.ndarray)):
        raise TypeError('pars must be list-like or numpy array')

    if len(pars) != 3:
        raise ValueError('pars must be of length 3')

    alpha, beta, gamma = pars
    return (1 - gamma) - (1 - 2 * gamma) * np.exp(-((xx / alpha) ** beta))


def weibull50(pars, xx):
    """
    Weibull function from 0.5 to 1, with lapse rate.
    Args:
        pars: Model parameters [alpha, beta, gamma].
        xx: vector of stim levels.
    Returns:
        A vector of length xx
    Raises:
        ValueError: pars must be of length 3
        TypeError: pars must be list-like or numpy array
    Information:
        2000-04 MC wrote it
        2018-08 MW ported to Python
    """
    # Validate input
    if not isinstance(pars, (list, tuple, np.ndarray)):
        raise TypeError('pars must be list-like or numpy array')

    if len(pars) != 3:
        raise ValueError('pars must be of length 3')

    alpha, beta, gamma = pars
    return (1 - gamma) - (.5 - gamma) * np.exp(-((xx / alpha) ** beta))


def erf_psycho(pars, xx):
    """
    erf function from 0 to 1, with lapse rate.
    Args:
        pars: Model parameters [bias, slope, lapse].
        xx: vector of stim levels.
    Returns:
        ff: A vector of length xx
    Examples:
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> xx = np.arange(-50, 50)
        >>> ff = erf_psycho(np.array([-10., 10., 0.1]), xx)
        >>> plt.plot(xx, ff)
    Raises:
        ValueError: pars must be of length 3
        TypeError: pars must be a list or numpy array
    Information:
        2000    MC wrote it
        2018-08 MW ported to Python
    """
    # Validate input
    if not isinstance(pars, (list, tuple, np.ndarray)):
        raise TypeError('pars must be list-like or numpy array')

    if len(pars) != 3:
        raise ValueError('pars must be of length 4')

    (bias, slope, gamma) = pars
    return gamma + (1 - 2 * gamma) * (erf((xx - bias) / slope) + 1) / 2


def erf_psycho_2gammas(pars, xx):
    """
    erf function from 0 to 1, with two lapse rates.
    Args:
        pars: Model parameters [bias, slope, gamma].
        xx: vector of stim levels (%)
    Returns:
        ff: A vector of length xx
    Examples:
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt
        >>> xx = np.arange(-50, 50)
        >>> ff = erf_psycho_2gammas(np.array([-10., 10., 0.2, 0.]), xx)
        >>> plt.plot(xx, ff)
    Raises:
        ValueError: pars must be of length 4
        TypeError: pars must be list-like or numpy array
    Information:
        2000    MC wrote it
        2018-08 MW ported to Python
    """
    # Validate input
    if not isinstance(pars, (list, tuple, np.ndarray)):
        raise TypeError('pars must be a list-like or numpy array')

    if len(pars) != 4:
        raise ValueError('pars must be of length 4')

    (bias, slope, gamma1, gamma2) = pars
    return gamma1 + (1 - gamma1 - gamma2) * (erf((xx - bias) / slope) + 1) / 2



class IblError(BaseException):
    explanation = ''

    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        return f"{self.message} \n {self.explanation} "


class SyncBpodWheelException(IblError):
    explanation = "The bpod can't be synchronized with the Rotary Encoder."


class SyncBpodFpgaException(IblError):
    explanation = "The bpod can't be synchronized with the FPGA."


class Neuropixel3BSyncFrontsNonMatching(IblError):
    explanation = (" When the npy files containing sync pulses for probes do not match with nidq."
                   "In 3B, this indicates that either the binary files is corrupt,"
                   "either the extracted sync files are corrupt.")


class AlyxSubjectNotFound(IblError):
    explanation = 'The subject was not found in Alyx database'



from collections import namedtuple
import ast


def read(filename):
    header = namedtuple('npy_header',
                        'magic_string, version, header_len, descr, fortran_order, shape')

    with open(filename, 'rb') as fid:
        header.magic_string = fid.read(6)
        header.version = fid.read(2)
        header.header_len = int.from_bytes(fid.read(2), byteorder='little')
        d = ast.literal_eval(fid.read(header.header_len).decode())

    for k in d.keys():
        print(k)
        setattr(header, k, d[k])

    return header



import shutil
import tempfile
import unittest
from pathlib import Path

import ibllib.io.raw_data_loaders as rawio
from ibllib.pipes import misc
from oneibl.one import ONE


class TestExtractors(unittest.TestCase):

    def test_task_names_extractors(self):
        """
        This is to test against regressions
        """
        task_out = [
            ('_iblrig_tasks_biasedChoiceWorld3.7.0', 'biased'),
            ('_iblrig_tasks_biasedScanningChoiceWorld5.2.3', 'biased'),
            ('_iblrig_tasks_trainingChoiceWorld3.6.0', 'training'),
            ('_iblrig_tasks_ephysChoiceWorld5.1.3', 'ephys'),
            ('_iblrig_calibration_frame2TTL4.1.3', None),
            ('_iblrig_tasks_habituationChoiceWorld3.6.0', 'habituation'),
            ('_iblrig_tasks_scanningOptoChoiceWorld5.0.0', None),
            ('_iblrig_tasks_RewardChoiceWorld4.1.3', None),
            ('_iblrig_calibration_screen4.1.3', None),
            ('_iblrig_tasks_ephys_certification4.1.3', 'sync_ephys'),
        ]
        for to in task_out:
            out = rawio.get_task_extractor_type(to[0])
            self.assertEqual(out, to[1])


class TestPipesMisc(unittest.TestCase):
    """
    """

    def setUp(self):
        # Define folders
        self.root_test_folder = Path(tempfile.gettempdir()) / 'ibllib_test'
        self.local_data_path = self.root_test_folder / 'local' / 'iblrig_data' / 'Subjects'
        self.remote_data_path = self.root_test_folder / 'remote' / 'iblrig_data' / 'Subjects'
        self.local_session_path_3A = self.local_data_path / \
            'test_subject' / '1900-01-01' / '001'
        self.local_session_path_3B = self.local_data_path / \
            'test_subject' / '1900-01-02' / '001'
        self.raw_ephys_data_path_3A = self.local_session_path_3A / 'raw_ephys_data'
        self.raw_ephys_data_path_3B = self.local_session_path_3B / 'raw_ephys_data'
        self.probe00_path_3A = self.raw_ephys_data_path_3A / 'probe00'
        self.probe01_path_3A = self.raw_ephys_data_path_3A / 'probe01'
        self.probe00_path_3B = self.raw_ephys_data_path_3B / 'probe00'
        self.probe01_path_3B = self.raw_ephys_data_path_3B / 'probe01'

        self.bad_ephys_folder_3A = self.raw_ephys_data_path_3A / 'bad_ephys_folder'
        self.bad_probe00_folder_3A = self.bad_ephys_folder_3A / 'some_probe00_folder'
        self.bad_probe01_folder_3A = self.bad_ephys_folder_3A / 'some_probe01_folder'

        self.bad_ephys_folder_3B = self.raw_ephys_data_path_3B / 'bad_folder_g0_t0'
        self.bad_probe00_folder_3B = self.bad_ephys_folder_3B / 'bad_folder_g0_t0.imec0'
        self.bad_probe01_folder_3B = self.bad_ephys_folder_3B / 'bad_folder_g0_t0.imec1'
        # Make folders
        (self.local_session_path_3A / 'raw_behavior_data').mkdir(exist_ok=True, parents=True)
        (self.local_session_path_3B / 'raw_behavior_data').mkdir(exist_ok=True, parents=True)
        self.probe00_path_3A.mkdir(exist_ok=True, parents=True)
        self.probe00_path_3B.mkdir(exist_ok=True, parents=True)
        self.probe01_path_3A.mkdir(exist_ok=True, parents=True)
        self.probe01_path_3B.mkdir(exist_ok=True, parents=True)

        self.bad_probe00_folder_3A.mkdir(exist_ok=True, parents=True)
        self.bad_probe01_folder_3A.mkdir(exist_ok=True, parents=True)
        self.bad_probe00_folder_3B.mkdir(exist_ok=True, parents=True)
        self.bad_probe01_folder_3B.mkdir(exist_ok=True, parents=True)
        # Make some files
        self.populate_bad_folders()

    def populate_bad_folders(self):
        # 3A
        (self.bad_probe00_folder_3A / 'blabla_g0_t0.imec.ap.bin').touch()
        (self.bad_probe00_folder_3A / 'blabla_g0_t0.imec.lf.bin').touch()
        (self.bad_probe00_folder_3A / 'blabla_g0_t0.imec.ap.meta').touch()
        (self.bad_probe00_folder_3A / 'blabla_g0_t0.imec.lf.meta').touch()

        (self.bad_probe01_folder_3A / 'blabla_g0_t0.imec.ap.bin').touch()
        (self.bad_probe01_folder_3A / 'blabla_g0_t0.imec.lf.bin').touch()
        (self.bad_probe01_folder_3A / 'blabla_g0_t0.imec.ap.meta').touch()
        (self.bad_probe01_folder_3A / 'blabla_g0_t0.imec.lf.meta').touch()
        # 3B
        (self.bad_probe00_folder_3B / 'blabla_g0_t0.imec0.ap.bin').touch()
        (self.bad_probe00_folder_3B / 'blabla_g0_t0.imec0.lf.bin').touch()
        (self.bad_probe00_folder_3B / 'blabla_g0_t0.imec0.ap.meta').touch()
        (self.bad_probe00_folder_3B / 'blabla_g0_t0.imec0.lf.meta').touch()

        (self.bad_probe01_folder_3B / 'blabla_g0_t0.imec1.ap.bin').touch()
        (self.bad_probe01_folder_3B / 'blabla_g0_t0.imec1.lf.bin').touch()
        (self.bad_probe01_folder_3B / 'blabla_g0_t0.imec1.ap.meta').touch()
        (self.bad_probe01_folder_3B / 'blabla_g0_t0.imec1.lf.meta').touch()

        (self.bad_ephys_folder_3B / 'blabla_g0_t0.nidq.bin').touch()
        (self.bad_ephys_folder_3B / 'blabla_g0_t0.nidq.meta').touch()

    def test_behavior_exists(self):
        assert(misc.behavior_exists('.') is False)
        assert(misc.behavior_exists(self.local_session_path_3A) is True)

    def test_check_transfer(self):
        misc.check_transfer(self.bad_probe00_folder_3A, self.bad_probe00_folder_3A)
        misc.check_transfer(str(self.bad_probe00_folder_3A), self.bad_probe00_folder_3A)
        with self.assertRaises(AssertionError) as cm:
            misc.check_transfer(self.bad_probe00_folder_3A, self.bad_probe00_folder_3B)
        err = cm.exception
        self.assertEqual(str(err), 'src_files != dst_files')

    def test_get_new_filename(self):
        binFname3A = 'ignoreThisPart_g0_t0.imec.ap.bin'
        binFname3B = 'ignoreThisPart_g0_t0.imec0.ap.bin'
        metaFname3A = 'ignoreThisPart_g0_t0.imec.ap.meta'
        metaFname3B = 'ignoreThisPart_g0_t0.imec0.ap.meta'
        probe1_3B_Meta = 'ignoreThisPart_g0_t0.imec1.ap.meta'
        different_gt = 'ignoreThisPart_g1_t2.imec0.ap.meta'

        newfname = misc.get_new_filename(binFname3A)
        self.assertTrue(newfname == '_spikeglx_ephysData_g0_t0.imec.ap.bin')
        newfname = misc.get_new_filename(binFname3B)
        self.assertTrue(newfname == '_spikeglx_ephysData_g0_t0.imec0.ap.bin')
        newfname = misc.get_new_filename(metaFname3A)
        self.assertTrue(newfname == '_spikeglx_ephysData_g0_t0.imec.ap.meta')
        newfname = misc.get_new_filename(metaFname3B)
        self.assertTrue(newfname == '_spikeglx_ephysData_g0_t0.imec0.ap.meta')
        newfname = misc.get_new_filename(probe1_3B_Meta)
        self.assertTrue(newfname == '_spikeglx_ephysData_g0_t0.imec1.ap.meta')
        newfname = misc.get_new_filename(different_gt)
        self.assertTrue(newfname == '_spikeglx_ephysData_g1_t2.imec0.ap.meta')

    def _test_rename_ephys_files(self):
        # Test 3A
        misc.rename_ephys_files(self.local_session_path_3A)
        ap_files = list(self.local_session_path_3A.rglob('*.ap.*'))
        lf_files = list(self.local_session_path_3A.rglob('*.lf.*'))
        self.assertTrue(len(ap_files) >= 1)
        self.assertTrue(len(lf_files) >= 1)
        for f in ap_files:
            print(f.parent.name, f.name)
            self.assertTrue('_spikeglx_ephysData_g' in str(f))
            self.assertTrue('_t' in str(f))
            self.assertTrue('.imec.ap' in str(f))
            self.assertTrue(('.bin' in str(f) or '.meta' in str(f)))
            self.assertTrue(('probe00' in str(f) or 'probe01' in str(f)))
        for f in lf_files:
            print(f.parent.name, f.name)
            self.assertTrue('_spikeglx_ephysData_g' in str(f))
            self.assertTrue('_t' in str(f))
            self.assertTrue('.imec.lf' in str(f))
            self.assertTrue(('.bin' in str(f) or '.meta' in str(f)))  # Test 3B
            self.assertTrue(('probe00' in str(f) or 'probe01' in str(f)))
        # Test 3B
        misc.rename_ephys_files(self.local_session_path_3B)
        ap_files = list(self.local_session_path_3B.rglob('*.ap.*'))
        lf_files = list(self.local_session_path_3B.rglob('*.lf.*'))
        nidq_files = list(self.local_session_path_3B.rglob('*.nidq.*'))
        for f in ap_files:
            print(f.name)
            self.assertTrue('_spikeglx_ephysData_g' in str(f))
            self.assertTrue('_t' in str(f))
            self.assertTrue(('.imec0.ap' in str(f) or 'imec1.ap' in str(f)))
            self.assertTrue(('.bin' in str(f) or '.meta' in str(f)))
        for f in lf_files:
            print(f.name)
            self.assertTrue('_spikeglx_ephysData_g' in str(f))
            self.assertTrue('_t' in str(f))
            self.assertTrue(('.imec0.lf' in str(f) or 'imec1.lf' in str(f)))
            self.assertTrue(('.bin' in str(f) or '.meta' in str(f)))  # Test 3B
        for f in nidq_files:
            print(f.name)
            self.assertTrue('_spikeglx_ephysData_g' in str(f))
            self.assertTrue('_t' in str(f))
            self.assertTrue('.nidq' in str(f))
            self.assertTrue(('.bin' in str(f) or '.meta' in str(f)))

    def test_rename_and_move(self):
        self._test_rename_ephys_files()
        # Test for 3A
        misc.move_ephys_files(self.local_session_path_3A)
        probe00_files = list(self.probe00_path_3A.rglob('*'))
        probe00_file_names = [x.name for x in probe00_files]
        probe01_files = list(self.probe01_path_3A.rglob('*'))
        probe01_file_names = [x.name for x in probe01_files]

        self.assertTrue(
            '_spikeglx_ephysData_g0_t0.imec.ap.bin' in probe00_file_names and
            '_spikeglx_ephysData_g0_t0.imec.ap.meta' in probe00_file_names and
            '_spikeglx_ephysData_g0_t0.imec.lf.bin' in probe00_file_names and
            '_spikeglx_ephysData_g0_t0.imec.lf.meta' in probe00_file_names
        )
        self.assertTrue(
            '_spikeglx_ephysData_g0_t0.imec.ap.bin' in probe01_file_names and
            '_spikeglx_ephysData_g0_t0.imec.ap.meta' in probe01_file_names and
            '_spikeglx_ephysData_g0_t0.imec.lf.bin' in probe01_file_names and
            '_spikeglx_ephysData_g0_t0.imec.lf.meta' in probe01_file_names
        )
        # Test for 3B
        misc.move_ephys_files(self.local_session_path_3B)
        probe00_files = list(self.probe00_path_3B.rglob('*'))
        probe00_file_names = [x.name for x in probe00_files]
        probe01_files = list(self.probe01_path_3B.rglob('*'))
        probe01_file_names = [x.name for x in probe01_files]
        nidq_files = list(self.raw_ephys_data_path_3B.glob('*.nidq.*'))
        nidq_file_names = [x.name for x in nidq_files]
        self.assertTrue(
            '_spikeglx_ephysData_g0_t0.imec0.ap.bin' in probe00_file_names and
            '_spikeglx_ephysData_g0_t0.imec0.ap.meta' in probe00_file_names and
            '_spikeglx_ephysData_g0_t0.imec0.lf.bin' in probe00_file_names and
            '_spikeglx_ephysData_g0_t0.imec0.lf.meta' in probe00_file_names
        )
        self.assertTrue(
            '_spikeglx_ephysData_g0_t0.imec1.ap.bin' in probe01_file_names and
            '_spikeglx_ephysData_g0_t0.imec1.ap.meta' in probe01_file_names and
            '_spikeglx_ephysData_g0_t0.imec1.lf.bin' in probe01_file_names and
            '_spikeglx_ephysData_g0_t0.imec1.lf.meta' in probe01_file_names
        )
        self.assertTrue(
            '_spikeglx_ephysData_g0_t0.nidq.bin' in nidq_file_names and
            '_spikeglx_ephysData_g0_t0.nidq.meta' in nidq_file_names
        )

    def test_create_ephys_transfer_done_flag(self):
        # Create ephys flag file for completed transfer
        misc.create_ephys_transfer_done_flag(self.local_session_path_3A)
        # Check it was created
        ephys = Path(self.local_session_path_3A).joinpath('ephys_data_transferred.flag')
        self.assertTrue(ephys.exists())
        # Remove it
        ephys.unlink()

    def test_create_video_transfer_done_flag(self):
        # Create video flag file for completed transfer
        misc.create_video_transfer_done_flag(self.local_session_path_3A)
        # Check it was created
        video = Path(self.local_session_path_3A).joinpath('video_data_transferred.flag')
        self.assertTrue(video.exists())
        # Remove it
        video.unlink()

    def test_check_create_raw_session_flag(self):
        raw_session = Path(self.local_session_path_3A).joinpath('raw_session.flag')
        ephys = Path(self.local_session_path_3A).joinpath('ephys_data_transferred.flag')
        video = Path(self.local_session_path_3A).joinpath('video_data_transferred.flag')
        # Check not created
        misc.check_create_raw_session_flag(self.local_session_path_3A)
        self.assertFalse(raw_session.exists())
        # Create only ephys flag
        misc.create_ephys_transfer_done_flag(self.local_session_path_3A)
        # Check not created
        misc.check_create_raw_session_flag(self.local_session_path_3A)
        self.assertFalse(raw_session.exists())
        ephys.unlink()
        # Create only video flag
        misc.create_video_transfer_done_flag(self.local_session_path_3A)
        # Check not created
        misc.check_create_raw_session_flag(self.local_session_path_3A)
        self.assertFalse(raw_session.exists())
        video.unlink()
        # Create ephys and video flag file for completed transfer
        misc.create_ephys_transfer_done_flag(self.local_session_path_3A)
        misc.create_video_transfer_done_flag(self.local_session_path_3A)
        # Check it was created
        misc.check_create_raw_session_flag(self.local_session_path_3A)
        self.assertTrue(raw_session.exists())
        # Check other flags deleted
        self.assertFalse(ephys.exists())
        self.assertFalse(video.exists())
        raw_session.unlink()

    def test_create_ephys_flags(self):
        extract = self.local_session_path_3A.joinpath('extract_ephys.flag')
        qc = self.local_session_path_3A.joinpath('raw_ephys_qc.flag')
        spike_sorting0 = self.local_session_path_3A / 'raw_ephys_data' / 'probe00'
        spike_sorting1 = self.local_session_path_3A / 'raw_ephys_data' / 'probe01'
        spike_sorting0 = spike_sorting0.joinpath('spike_sorting.flag')
        spike_sorting1 = spike_sorting1.joinpath('spike_sorting.flag')
        misc.create_ephys_flags(self.local_session_path_3A)
        self.assertTrue(extract.exists())
        self.assertTrue(qc.exists())
        self.assertTrue(spike_sorting0.exists())
        self.assertTrue(spike_sorting1.exists())
        # Test recreate
        misc.create_ephys_flags(self.local_session_path_3A)
        self.assertTrue(extract.exists())
        self.assertTrue(qc.exists())
        self.assertTrue(spike_sorting0.exists())
        self.assertTrue(spike_sorting1.exists())
        # Remove flags after test
        extract.unlink()
        qc.unlink()
        spike_sorting0.unlink()
        spike_sorting1.unlink()
        # test removal
        self.assertFalse(extract.exists())
        self.assertFalse(qc.exists())
        self.assertFalse(spike_sorting0.exists())
        self.assertFalse(spike_sorting1.exists())

    def test_create_alyx_probe_insertions(self):
        # Connect to test DB
        one = ONE(
            username='test_user',
            password='TapetesBloc18',
            base_url='https://test.alyx.internationalbrainlab.org'
        )
        # Use existing session on test database
        eid = 'b1c968ad-4874-468d-b2e4-5ffa9b9964e9'
        # Force probe insertion 3A
        misc.create_alyx_probe_insertions(
            eid,
            one=one,
            model='3A',
            labels=['probe00', 'probe01'])
        # Verify it's been inserted
        alyx_insertion = one.alyx.rest('insertions', 'list',
                                       session=eid)
        self.assertTrue(alyx_insertion[0]['model'] == '3A')
        self.assertTrue(alyx_insertion[0]['name'] in ['probe00', 'probe01'])
        self.assertTrue(alyx_insertion[1]['model'] == '3A')
        self.assertTrue(alyx_insertion[1]['name'] in ['probe00', 'probe01'])
        # Cleanup DB
        one.alyx.rest('insertions', 'delete', id=alyx_insertion[0]['id'])
        one.alyx.rest('insertions', 'delete', id=alyx_insertion[1]['id'])
        # Force probe insertion 3B
        misc.create_alyx_probe_insertions(
            eid,
            one=one,
            model='3B2',
            labels=['probe00', 'probe01'])
        # Verify it's been inserted
        alyx_insertion = one.alyx.rest('insertions', 'list',
                                       session=eid)
        self.assertTrue(alyx_insertion[0]['model'] == '3B2')
        self.assertTrue(alyx_insertion[0]['name'] in ['probe00', 'probe01'])
        self.assertTrue(alyx_insertion[1]['model'] == '3B2')
        self.assertTrue(alyx_insertion[1]['name'] in ['probe00', 'probe01'])
        # Cleanup DB
        one.alyx.rest('insertions', 'delete', id=alyx_insertion[0]['id'])
        one.alyx.rest('insertions', 'delete', id=alyx_insertion[1]['id'])

    def tearDown(self):
        shutil.rmtree(self.root_test_folder, ignore_errors=True)


if __name__ == "__main__":
    unittest.main(exit=False)



import unittest

from ibllib.qc.oneqc_metrics import ONEQC
from oneibl.one import ONE


one = ONE(base_url='https://test.alyx.internationalbrainlab.org', username='test_user',
          password='TapetesBloc18')


class TestONEQCMetrics(unittest.TestCase):
    """
    Testing of one dstypes will require test sessions with one data in the test db
    """
    def setUp(self):
        # An ephys session on the test DB
        # Subj from testDB = clns0730
        self.one = one
        self.test_eid = 'cf264653-2deb-44cb-aa84-89b82507028a'

    def test_ONEQC_lazy(self):
        oneqc = ONEQC(self.test_eid, one=self.one, bpod_ntrials=None, lazy=True)
        self.assertTrue(oneqc.bpod_ntrials == 616)
        self.assertTrue(oneqc.metrics is None)
        self.assertTrue(oneqc.passed is None)
        oneqc.compute()
        self.assertTrue(oneqc.metrics is not None)
        self.assertTrue(oneqc.passed is not None)

    def test_ONEQC(self):
        oneqc = ONEQC(self.test_eid, one=self.one, bpod_ntrials=None, lazy=False)
        self.assertTrue(oneqc.metrics is not None)
        self.assertTrue(oneqc.passed is not None)


if __name__ == "__main__":
    unittest.main(exit=False)



from brainbox.singlecell import acorr, calculate_peths
import unittest
import numpy as np


class TestPopulation(unittest.TestCase):
    def test_acorr_0(self):
        spike_times = np.array([0, 10, 10, 20])
        bin_size = 1
        winsize_bins = 2 * 3 + 1

        c_expected = np.zeros(7, dtype=np.int32)
        c_expected[3] = 1

        c = acorr(spike_times, bin_size=bin_size, window_size=winsize_bins)

        self.assertTrue(np.allclose(c, c_expected))

    def test_acorr_1(self):
        spike_times = np.array([0, 10, 10, 20], dtype=np.float64)
        bin_size = 1
        winsize_bins = 2 * 3 + 1

        c_expected = np.zeros(7, dtype=np.float64)
        c_expected[3] = 1

        c = acorr(spike_times, bin_size=bin_size, window_size=winsize_bins)

        self.assertTrue(np.allclose(c, c_expected))


class TestPeths(unittest.TestCase):
    def test_peths_synthetic(self):
        n_spikes = 20000
        n_clusters = 20
        n_events = 200
        record_length = 1654
        cluster_sel = [1, 2, 3, 6, 15, 16]
        np.random.seed(seed=42)
        spike_times = np.sort(np.random.rand(n_spikes, ) * record_length)
        spike_clusters = np.random.randint(0, n_clusters, n_spikes)
        event_times = np.sort(np.random.rand(n_events, ) * record_length)

        peth, fr = calculate_peths(spike_times, spike_clusters, cluster_ids=cluster_sel,
                                   align_times=event_times)
        self.assertTrue(peth.means.shape[0] == len(cluster_sel))
        self.assertTrue(np.all(peth.means.shape == peth.stds.shape))
        self.assertTrue(np.all(fr.shape == (n_events, len(cluster_sel), 28)))
        self.assertTrue(peth.tscale.size == 28)


def test_firing_rate():
    pass


if __name__ == "__main__":
    np.random.seed(0)
    unittest.main(exit=False)



from pathlib import Path
import logging

import matplotlib.pyplot as plt
import numpy as np

from ibllib.ephys.neuropixel import SITES_COORDINATES
from oneibl.one import ONE
import alf.io
import ibllib.atlas as atlas
from ibllib.ephys.spikes import probes_description as extract_probes
from ibllib.dsp.utils import fcn_cosine
from ibllib.ephys.neuropixel import TIP_SIZE_UM


_logger = logging.getLogger('ibllib')

# origin Allen left, front, up
brain_atlas = atlas.AllenAtlas(res_um=25)


def load_track_csv(file_track):
    """
    Loads a lasagna track and convert to IBL-ALlen coordinate framework
    :param file_track:
    :return: xyz
    """
    # apmldv in the histology file is flipped along y direction
    file_track = Path(file_track)
    if file_track.stat().st_size == 0:
        return np.array([])
    ixiyiz = np.loadtxt(file_track, delimiter=',')[:, [1, 0, 2]]
    ixiyiz[:, 1] = 527 - ixiyiz[:, 1]
    ixiyiz = ixiyiz[np.argsort(ixiyiz[:, 2]), :]
    xyz = brain_atlas.bc.i2xyz(ixiyiz)
    # xyz[:, 0] = - xyz[:, 0]
    return xyz


def get_picked_tracks(histology_path, glob_pattern="*_pts_transformed.csv"):
    """
    This outputs reads in the Lasagna output and converts the picked tracks in the IBL coordinates
    :param histology_path: Path object: folder path containing tracks
    :return: xyz coordinates in
    """
    xyzs = []
    histology_path = Path(histology_path)
    if histology_path.is_file():
        files_track = [histology_path]
    else:
        files_track = list(histology_path.rglob(glob_pattern))
    for file_track in files_track:
        xyzs.append(load_track_csv(file_track))
    return {'files': files_track, 'xyz': xyzs}


def get_micro_manipulator_data(subject, one=None, force_extract=False):
    """
    Looks for all ephys sessions for a given subject and get the probe micro-manipulator
    trajectories.
    If probes ALF object not on flat-iron, attempts to perform the extraction from meta-data
    and task settings file.
    """
    if not one:
        one = ONE()

    eids, sessions = one.search(subject=subject, task_protocol='ephys', details=True)
    probes = alf.io.AlfBunch({})
    for ses in sessions:
        sess_path = Path(ses['local_path'])
        probe = None
        if not force_extract:
            probe = one.load_object(ses['url'], 'probes')
        if not probe:
            _logger.warning(f"Re-extraction probe info for {sess_path}")
            dtypes = ['_iblrig_taskSettings.raw', 'ephysData.raw.meta']
            raw_files = one.load(ses['url'], dataset_types=dtypes, download_only=True)
            if all([rf is None for rf in raw_files]):
                _logger.warning(f"no raw settings files nor ephys data found for"
                                f" {ses['local_path']}. Skip this session.")
                continue
            extract_probes(sess_path, bin_exists=False)
            probe = alf.io.load_object(sess_path.joinpath('alf'), 'probes')
        one.load(ses['url'], dataset_types='channels.localCoordinates', download_only=True)
        # get for each insertion the sites local mapping: if not found assumes checkerboard pattern
        probe['sites_coordinates'] = []
        for prb in probe.description:
            chfile = Path(ses['local_path']).joinpath('alf', prb['label'],
                                                      'channels.localCoordinates.npy')
            if chfile.exists():
                probe['sites_coordinates'].append(np.load(chfile))
            else:
                _logger.warning(f"no channel.localCoordinates found for {ses['local_path']}."
                                f"Assumes checkerboard pattern")
                probe['sites_coordinates'].append(SITES_COORDINATES)
        # put the session information in there
        probe['session'] = [ses] * len(probe.description)
        probes = probes.append(probe)
    return probes


def plot2d_all(trajectories, tracks):
    """
    Plot all tracks on a single 2d slice
    :param trajectories: dictionary output of the Alyx REST query on trajectories
    :param tracks:
    :return:
    """
    plt.figure()
    axs = brain_atlas.plot_sslice(brain_atlas.bc.i2x(190), cmap=plt.get_cmap('bone'))
    plt.figure()
    axc = brain_atlas.plot_cslice(brain_atlas.bc.i2y(350))
    for xyz in tracks['xyz']:
        axc.plot(xyz[:, 0] * 1e3, xyz[:, 2] * 1e6, 'b')
        axs.plot(xyz[:, 1] * 1e3, xyz[:, 2] * 1e6, 'b')
    for trj in trajectories:
        ins = atlas.Insertion.from_dict(trj, brain_atlas=brain_atlas)
        xyz = ins.xyz
        axc.plot(xyz[:, 0] * 1e3, xyz[:, 2] * 1e6, 'r')
        axs.plot(xyz[:, 1] * 1e3, xyz[:, 2] * 1e6, 'r')


def plot3d_all(trajectories, tracks):
    """
    Plot all tracks on a single 2d slice
    :param trajectories: dictionary output of the Alyx REST query on trajectories
    :param tracks:
    :return:
    """
    from mayavi import mlab
    src = mlab.pipeline.scalar_field(brain_atlas.label)
    mlab.pipeline.iso_surface(src, contours=[0.5, ], opacity=0.3)

    pts = []
    for xyz in tracks['xyz']:
        mlapdv = brain_atlas.bc.xyz2i(xyz)
        pts.append(mlab.plot3d(mlapdv[:, 1], mlapdv[:, 0], mlapdv[:, 2], line_width=3))

    plt_trj = []
    for trj in trajectories:
        ins = atlas.Insertion.from_dict(trj, brain_atlas=brain_atlas)
        mlapdv = brain_atlas.bc.xyz2i(ins.xyz)
        plt = mlab.plot3d(mlapdv[:, 1], mlapdv[:, 0], mlapdv[:, 2],
                          line_width=3, color=(1., 0., 1.))
        plt_trj.append(plt)


def interpolate_along_track(xyz_track, depths):
    """
    Get the coordinates of points along a track according to their distances from the first
    point.
    :param xyz_track: np.array [npoints, 3]. Usually the first point is the deepest
    :param depths: distance from the first point of the track, usually the convention is the
    deepest point is 0 and going up
    :return: xyz_channels
    """
    # from scipy.interpolate import interp1d
    # this is the cumulative distance from the lowest picked point (first)
    distance = np.cumsum(np.r_[0, np.sqrt(np.sum(np.diff(xyz_track, axis=0) ** 2, axis=1))])
    xyz_channels = np.zeros((depths.shape[0], 3))
    for m in np.arange(3):
        xyz_channels[:, m] = np.interp(depths, distance, xyz_track[:, m])
        # xyz_channels[:, m] = interp1d(distance, xyz[:, m], kind='cubic')(chdepths / 1e6)
    # plt.figure()
    # plt.plot(xyz_track[:, 0] * 1e6, xyz_track[:, 2] * 1e6, 'k*'), plt.axis('equal')
    # plt.plot(xyz_channels[:, 0] * 1e6, xyz_channels[:, 2] * 1e6, '.'), plt.axis('equal')
    return xyz_channels


def get_brain_regions(xyz, channels_positions=SITES_COORDINATES, brain_atlas=brain_atlas):
    """
    :param xyz: numpy array of 3D coordinates corresponding to a picked track or a trajectory
    the deepest point is assumed to be the tip.
    :param channels_positions:
    :param brain_atlas:
    :return: brain_regions (associated to each channel),
             insertion (object atlas.Insertion, defining 2 points of entries
             (tip and end of probe))
    """
    """
    this is the depth along the probe (from the first point which is the deepest labeled point)
    Due to the blockiness, depths may not be unique along the track so it has to be prepared
    """
    xyz = xyz[np.argsort(xyz[:, 2]), :]
    d = atlas.cart2sph(xyz[:, 0] - xyz[0, 0], xyz[:, 1] - xyz[0, 1], xyz[:, 2] - xyz[0, 2])[0]
    indsort = np.argsort(d)
    xyz = xyz[indsort, :]
    d = d[indsort]
    iduplicates = np.where(np.diff(d) == 0)[0]
    xyz = np.delete(xyz, iduplicates, axis=0)
    d = np.delete(d, iduplicates, axis=0)

    assert np.all(np.diff(d) > 0), "Depths should be strictly increasing"

    # Get the probe insertion from the coordinates
    insertion = atlas.Insertion.from_track(xyz, brain_atlas)

    # Interpolate channel positions along the probe depth and get brain locations
    TIP_SIZE_UM = 200
    xyz_channels = interpolate_along_track(xyz, (channels_positions[:, 1] + TIP_SIZE_UM) / 1e6)

    # get the brain regions
    brain_regions = brain_atlas.regions.get(brain_atlas.get_labels(xyz_channels))
    brain_regions['xyz'] = xyz_channels
    brain_regions['lateral'] = channels_positions[:, 0]
    brain_regions['axial'] = channels_positions[:, 1]
    assert np.unique([len(brain_regions[k]) for k in brain_regions]).size == 1

    return brain_regions, insertion


def register_track(probe_id, picks=None, one=None, overwrite=False):
    """
    Register the user picks to a probe in Alyx
    Here we update Alyx models on the database in 3 steps
    1) The user picks converted to IBL coordinates will be stored in the json field of the
    corresponding probe insertion models
    2) The trajectory computed from the histology track is created or patched
    3) Channel locations are set in the table
    """
    assert one
    # 0) if it's an empty track, create a null trajectory and exit
    if picks is None or picks.size == 0:
        tdict = {'probe_insertion': probe_id,
                 'x': None, 'y': None, 'z': None,
                 'phi': None, 'theta': None, 'depth': None, 'roll': None,
                 'provenance': 'Histology track',
                 'coordinate_system': 'IBL-Allen',
                 }
        brain_locations = None
    else:
        brain_locations, insertion_histology = get_brain_regions(picks)
        # 1) update the alyx models, first put the picked points in the insertion json
        one.alyx.rest('insertions', 'partial_update',
                      id=probe_id,
                      data={'json': {'xyz_picks': np.int32(picks * 1e6).tolist()}})
        # 2) patch or create the trajectory coming from histology track
        tdict = create_trajectory_dict(probe_id, insertion_histology, provenance='Histology track')

    hist_traj = one.alyx.rest('trajectories', 'list',
                              probe_insertion=probe_id,
                              provenance='Histology track')
    # if the trajectory exists, remove it, this will cascade delete existing channel locations
    if len(hist_traj):
        if overwrite:
            one.alyx.rest('trajectories', 'delete', id=hist_traj[0]['id'])
        else:
            raise FileExistsError('The session already exists, however overwrite is set to False.'
                                  'If you want to overwrite, set overwrite=True.')
    hist_traj = one.alyx.rest('trajectories', 'create', data=tdict)

    if brain_locations is None:
        return brain_locations, None
    # 3) create channel locations
    channel_dict = create_channel_dict(hist_traj, brain_locations)
    one.alyx.rest('channels', 'create', data=channel_dict)
    return brain_locations, insertion_histology


def register_aligned_track(probe_id, insertion, brain_locations, one=None, overwrite=False):
    """
    Register ephys aligned trajectory and channel locations to Alyx
    Here we update Alyx models on the database in 2 steps
    1) The trajectory computed from the final electrode channel locations
    2) Channel locations are set to the trajectory
    """
    assert one
    tdict = create_trajectory_dict(probe_id, insertion, provenance='Ephys aligned histology track')

    hist_traj = one.alyx.rest('trajectories', 'list',
                              probe_insertion=probe_id,
                              provenance='Ephys aligned histology track')
    # if the trajectory exists, remove it, this will cascade delete existing channel locations
    if len(hist_traj):
        if overwrite:
            one.alyx.rest('trajectories', 'delete', id=hist_traj[0]['id'])
        else:
            raise FileExistsError('The session already exists, however overwrite is set to False.'
                                  'If you want to overwrite, set overwrite=True.')
    hist_traj = one.alyx.rest('trajectories', 'create', data=tdict)

    channel_dict = create_channel_dict(hist_traj, brain_locations)
    one.alyx.rest('channels', 'create', data=channel_dict)


def create_trajectory_dict(probe_id, insertion, provenance):
    """
    Create trajectory dictionary in form to upload to alyx
    :param probe id: unique id of probe insertion
    :type probe_id: string (hexadecimal UUID)
    :param insertion: Insertion object describing entry and tip of trajectory
    :type insertion: object atlas.Insertion
    :param provenance: 'Histology track' or 'Ephys aligned histology track'
    :type provenance: string
    :return tdict:
    :type tdict: dict
    """
    tdict = {'probe_insertion': probe_id,
             'x': insertion.x * 1e6,
             'y': insertion.y * 1e6,
             'z': insertion.z * 1e6,
             'phi': insertion.phi,
             'theta': insertion.theta,
             'depth': insertion.depth * 1e6,
             'roll': insertion.beta,
             'provenance': provenance,
             'coordinate_system': 'IBL-Allen',
             }

    return tdict


def create_channel_dict(traj, brain_locations):
    """
    Create channel dictionary in form to upload to alyx
    :param traj: alyx trajectory object to attach channel information to
    :type traj: dict
    :param brain_locations: information about location of electrode channels in brain atlas
    :type insertion: Bunch
    :return tdict:
    :type tdict: list of dict
    """
    channel_dict = []
    for i in np.arange(brain_locations.id.size):
        channel_dict.append({
            'x': brain_locations.xyz[i, 0] * 1e6,
            'y': brain_locations.xyz[i, 1] * 1e6,
            'z': brain_locations.xyz[i, 2] * 1e6,
            'axial': brain_locations.axial[i],
            'lateral': brain_locations.lateral[i],
            'brain_region': int(brain_locations.id[i]),
            'trajectory_estimate': traj['id']
        })

    return channel_dict


def _parse_filename(track_file):
    tmp = track_file.name.split('_')
    inumber = [i for i, s in enumerate(tmp) if s.isdigit and len(s) == 3][-1]
    search_filter = {'date': tmp[0], 'experiment_number': int(tmp[inumber]),
                     'name': '_'.join(tmp[inumber + 1:- 1]),
                     'subject': '_'.join(tmp[1:inumber])}
    return search_filter


def register_track_files(path_tracks, one=None, overwrite=False):
    """
    :param path_tracks: path to directory containing tracks; also works with a single file name
    :param one:
    :return:
    """
    glob_pattern = "*_probe*_pts*.csv"

    path_tracks = Path(path_tracks)

    if not path_tracks.is_dir():
        track_files = [path_tracks]
    else:
        track_files = list(path_tracks.rglob(glob_pattern))
        track_files.sort()

    assert path_tracks.exists()
    assert one

    ntracks = len(track_files)
    for ind, track_file in enumerate(track_files):
        # Nomenclature expected:
        # '{yyyy-mm-dd}}_{nickname}_{session_number}_{probe_label}_pts.csv'
        # beware: there may be underscores in the subject nickname

        search_filter = _parse_filename(track_file)
        probe = one.alyx.rest('insertions', 'list', **search_filter)
        if len(probe) == 0:
            eid = one.search(subject=search_filter['subject'], date_range=search_filter['date'],
                             number=search_filter['experiment_number'])
            if len(eid) == 0:
                raise Exception(f"No session found {track_file.name}")
            insertion = {'session': eid[0],
                         'name': search_filter['name']}
            probe = one.alyx.rest('insertions', 'create', data=insertion)
        elif len(probe) == 1:
            probe = probe[0]
        else:
            raise ValueError("Multiple probes found.")
        probe_id = probe['id']
        try:
            xyz_picks = load_track_csv(track_file)
            register_track(probe_id, xyz_picks, one=one, overwrite=overwrite)
        except Exception as e:
            _logger.error(str(track_file))
            raise e
        _logger.info(f"{ind + 1}/{ntracks}, {str(track_file)}")


def coverage(trajs, ba=None):
    """
    Computes a coverage volume from
    :param trajs: dictionary of trajectories from Alyx rest endpoint (one.alyx.rest...)
    :param ba: ibllib.atlas.BrainAtlas instance
    :return: 3D np.array the same size as the volume provided in the brain atlas
    """
    # in um. Coverage = 1 below the first value, 0 after the second, cosine taper in between
    DIST_FCN = [100, 150]
    ACTIVE_LENGTH_UM = 3.5 * 1e3
    MAX_DIST_UM = DIST_FCN[1]  # max distance around the probe to be searched for
    if ba is None:
        ba = atlas.AllenAtlas()

    def crawl_up_from_tip(ins, d):
        return (ins.entry - ins.tip) * (d[:, np.newaxis] /
                                        np.linalg.norm(ins.entry - ins.tip)) + ins.tip

    full_coverage = np.zeros(ba.image.shape, dtype=np.float32).flatten()

    for p in np.arange(len(trajs)):
        if p % 20 == 0:
            print(p / len(trajs))
        traj = trajs[p]

        ins = atlas.Insertion.from_dict(traj)
        # those are the top and bottom coordinates of the active part of the shank extended
        # to maxdist
        d = (np.array([ACTIVE_LENGTH_UM + MAX_DIST_UM * np.sqrt(2),
                       -MAX_DIST_UM * np.sqrt(2)]) + TIP_SIZE_UM)
        top_bottom = crawl_up_from_tip(ins, d / 1e6)
        # this is the axis that has the biggest deviation. Almost always z
        axis = np.argmax(np.abs(np.diff(top_bottom, axis=0)))
        if axis != 2:
            _logger.warning(f"This works only for 45 degree or vertical tracks so far, skipping"
                            f" {ins}")
            continue
        # sample the active track path along this axis
        tbi = ba.bc.xyz2i(top_bottom)
        nz = tbi[1, axis] - tbi[0, axis] + 1
        ishank = np.round(np.array(
            [np.linspace(tbi[0, i], tbi[1, i], nz) for i in np.arange(3)]).T).astype(np.int32)

        # creates a flattened "column" of candidate volume indices around the track
        # around each sample get an horizontal square slice of nx *2 +1 and ny *2 +1 samples
        # flatten the corresponding xyz indices and  compute the min distance to the track only
        # for those
        nx = int(np.floor(MAX_DIST_UM / 1e6 / np.abs(ba.bc.dxyz[0]) * np.sqrt(2) / 2)) * 2 + 1
        ny = int(np.floor(MAX_DIST_UM / 1e6 / np.abs(ba.bc.dxyz[1]) * np.sqrt(2) / 2)) * 2 + 1
        ixyz = np.stack([v.flatten() for v in np.meshgrid(
            np.arange(-nx, nx + 1), np.arange(-ny, ny + 1), np.arange(nz))]).T
        ixyz[:, 0] = ishank[ixyz[:, 2], 0] + ixyz[:, 0]
        ixyz[:, 1] = ishank[ixyz[:, 2], 1] + ixyz[:, 1]
        ixyz[:, 2] = ishank[ixyz[:, 2], 2]
        # if any, remove indices that lie outside of the volume bounds
        iok = np.logical_and(0 <= ixyz[:, 0], ixyz[:, 0] < ba.bc.nx)
        iok &= np.logical_and(0 <= ixyz[:, 1], ixyz[:, 1] < ba.bc.ny)
        iok &= np.logical_and(0 <= ixyz[:, 2], ixyz[:, 2] < ba.bc.nz)
        ixyz = ixyz[iok, :]
        # get the minimum distance to the trajectory, to which is applied the cosine taper
        xyz = np.c_[ba.bc.xscale[ixyz[:, 0]], ba.bc.yscale[ixyz[:, 1]], ba.bc.zscale[ixyz[:, 2]]]
        sites_bounds = crawl_up_from_tip(
            ins, (np.array([ACTIVE_LENGTH_UM, 0]) + TIP_SIZE_UM) / 1e6)
        mdist = ins.trajectory.mindist(xyz, bounds=sites_bounds)
        coverage = 1 - fcn_cosine(np.array(DIST_FCN) / 1e6)(mdist)
        # remap to the coverage volume
        full_coverage[ba._lookup_inds(ixyz)] += coverage

    full_coverage = full_coverage.reshape(ba.image.shape)
    full_coverage[ba.label == 0] = np.nan
    return full_coverage



"""
Quality control for arbitrary metrics, using permutation testing.

Written by Sebastian Bruijns
"""

import numpy as np
import time
import matplotlib.pyplot as plt
# TODO: take in eids and download data yourself?


def permut_test(data1, data2, metric, n_permut=1000, show=False, title=None):
    """
    Compute the probability of observating metric difference for datasets, via permutation testing.

    We're taking absolute values of differences, because the order of dataset input shouldn't
    matter
    We're only computing means, what if we want to apply a more complicated function to the
    permutation result?
    Pay attention to always give one list (even if its just one dataset, but then it doesn't make
    sense anyway...)

    Parameters
    ----------
    data1 : array-like
        First data set, list or array of data-entities to use for permutation test
        (make data2 optional and then permutation test more similar to tuning sensitivity?)
    data2 : array-like
        Second data set, also list or array of data-entities to use for permutation test
    metric : function, array-like -> float
        Metric to use for permutation test, will be used to reduce elements of data1 and data2
        to one number
    n_permut : integer (optional)
        Number of perumtations to use for test
    plot : Boolean (optional)
        Whether or not to show a plot of the permutation distribution and a marker for the position
        of the true difference in relation to this distribution

    Returns
    -------
    p : float
        p-value of true difference in permutation distribution

    See Also
    --------
    TODO:

    Examples
    --------
    TODO:
    """
    # Calculate metrics and true difference between groups
    print('data1')
    print(data1)
    metrics1 = [metric(d) for d in data1]
    print('metrics1')
    print(metrics1)
    metrics2 = [metric(d) for d in data2]
    true_diff = np.abs(np.mean(metrics1) - np.mean(metrics2))

    # Prepare permutations
    size1 = len(metrics1)
    diffs = np.concatenate((metrics1, metrics2))
    permutations = np.zeros((n_permut, diffs.size), dtype=np.int32)

    # Create permutations, could be parallelized or vectorized in principle, but unclear how
    indizes = np.arange(diffs.size)
    for i in range(n_permut):
        np.random.shuffle(indizes)
        permutations[i] = indizes

    permut_diffs = np.abs(np.mean(diffs[permutations[:, :size1]], axis=1) -
                          np.mean(diffs[permutations[:, size1:]], axis=1))
    p = len(permut_diffs[permut_diffs > true_diff]) / n_permut

    if show or title:
        plot_permut_test(permut_diffs=permut_diffs, true_diff=true_diff, p=p, title=title)

    return p


def plot_permut_test(permut_diffs, true_diff, p, title=None):
    """Plot permutation test result."""
    n, _, _ = plt.hist(permut_diffs)
    plt.plot(true_diff, np.max(n) / 20, '*r', markersize=12)

    # Prettify plot
    plt.gca().spines['top'].set_visible(False)
    plt.gca().spines['right'].set_visible(False)
    plt.title("p = {}".format(p))

    if title:
        plt.savefig(title + '.png')
    plt.close()


if __name__ == '__main__':
    rng = np.random.RandomState(2)
    data1 = rng.normal(0, 1, (23, 5))
    data2 = rng.normal(0.1, 1, (32, 5))
    t = time.time()
    p = permut_test(data1, data2, np.mean, plot=True)
    print(time.time() - t)
    print(p)



import hashlib
from pathlib import Path

import numpy as np
from tqdm import tqdm

BUF_SIZE = 2 ** 28  # 256 megs


def md5(file_path):
    """
    Computes md5 hash in a memory reasoned way
    md5hash = hashfile.md5(file_path)
    """
    return _hash_file(file_path, hashlib.md5())


def sha1(file_path):
    """
    Computes sha1 hash in a memory reasoned way
    md5hash = hashfile.sha1(file_path)
    """
    return _hash_file(file_path, hashlib.sha1())


def _hash_file(file_path, hash_obj, progress_bar=None):
    file_path = Path(file_path)
    file_size = file_path.stat().st_size
    # by default prints a progress bar only for files above 512 Mo
    if progress_bar is None:
        progress_bar = file_size > (512 * 1024 * 1024)
    b = bytearray(BUF_SIZE)
    mv = memoryview(b)
    pbar = tqdm(total=np.ceil(file_size / BUF_SIZE), disable=not progress_bar)
    with open(file_path, 'rb', buffering=0) as f:
        for n in iter(lambda: f.readinto(mv), 0):
            hash_obj.update(mv[:n])
            pbar.update(1)
    pbar.close()
    return hash_obj.hexdigest()



import unittest
from pathlib import Path
from oneibl.one import ONE

one = ONE(
    base_url="https://test.alyx.internationalbrainlab.org",
    username="test_user",
    password="TapetesBloc18",
)


class TestUtils(unittest.TestCase):
    def setUp(self):
        # Init connection to the database
        self.eids = [
            "cf264653-2deb-44cb-aa84-89b82507028a",
            "4e0b3320-47b7-416e-b842-c34dc9004cf8",
            "a9d89baf-9905-470c-8565-859ff212c7be",
            "aaf101c3-2581-450a-8abd-ddb8f557a5ad",
        ]
        self.partial_eid_paths = [
            Path('FlatIron/mainenlab/Subjects/clns0730/2018-08-24/002'),
            Path('FlatIron/zadorlab/Subjects/flowers/2018-07-13/001'),
            Path("FlatIron/mainenlab/Subjects/ZM_1743/2019-06-04/001"),
            Path("FlatIron/cortexlab/Subjects/KS005/2019-04-04/004"),
        ]
        self.det_keys = [
            "subject",
            "start_time",
            "number",
            "lab",
            "project",
            "url",
            "task_protocol",
            "local_path",
        ]
        self.full_det_keys = [
            "subject",
            "users",
            "location",
            "procedures",
            "lab",
            "project",
            "type",
            "task_protocol",
            "number",
            "start_time",
            "end_time",
            "narrative",
            "parent_session",
            "n_correct_trials",
            "n_trials",
            "url",
            "extended_qc",
            "qc",
            "wateradmin_session_related",
            "data_dataset_session_related",
            "json",
            "probe_insertion",
        ]

    def test_path_from_eid(self):
        # Test if eid's produce correct output
        for e, p in zip(self.eids, self.partial_eid_paths):
            self.assertTrue(str(p) in str(one.path_from_eid(e)))
        # Test if list input produces valid list output
        list_output = one.path_from_eid(self.eids)
        self.assertTrue(isinstance(list_output, list))
        self.assertTrue(
            all([str(p) in str(o) for p, o in zip(self.partial_eid_paths, list_output)])
        )

    def test_eid_from_path(self):
        # test if paths produce expected eid's
        paths = self.partial_eid_paths[-2:]
        paths.append("FlatIron/mainenlab/Subjects/ZM_1743/2019-06-04/001/bla.ble")
        paths.append(
            "some/other/root/FlatIron/cortexlab/" "Subjects/KS005/2019-04-04/004/bli/blo.blu"
        )
        eids = self.eids[-2:]
        eids.append("a9d89baf-9905-470c-8565-859ff212c7be")
        eids.append("aaf101c3-2581-450a-8abd-ddb8f557a5ad")
        for p, e in zip(paths, eids):
            self.assertTrue(e == str(one.eid_from_path(p)))
        # Test if list input produces correct list output
        list_output = one.eid_from_path(paths)
        self.assertTrue(isinstance(list_output, list))
        self.assertTrue(all([e == o for e, o in zip(eids, list_output)]))

    def test_get_details(self):
        # Get one known eid details dics
        det_from_eid = one.get_details(self.eids[0])
        full_det_from_eid = one.get_details(self.eids[0], full=True)
        # Test if all keys in returned dict are in expected keys
        # Full dict output
        self.assertTrue(
            all([x in full_det_from_eid for x in self.full_det_keys]),
            "Missing details key"
        )
        # Partial dict output
        self.assertTrue(
            all([x in det_from_eid for x in self.det_keys]),
            "Missing details key"
        )
        # Test local path append to details
        self.assertTrue(det_from_eid['local_path'] is not None)
        # Test is known local_path is not None
        self.assertTrue(one.get_details(self.eids[2])['local_path'] is not None)
        # Test list input
        self.assertTrue(len(self.eids) == len(one.get_details(self.eids)))


if __name__ == "__main__":
    unittest.main(exit=False)
